from pathlib import Path

from sanic_ext.extensions.openapi.builders import SpecificationBuilder


def test_specification_singleton():
    spec = SpecificationBuilder()
    spec.tag("foo")

    new_spec = SpecificationBuilder()

    assert spec is new_spec
    assert spec.tags == new_spec.tags


def test_specification_reset(get_docs):
    spec = SpecificationBuilder()

    docs = get_docs()
    assert len(docs["tags"]) == 0

    spec.tag("foo")
    docs = get_docs()
    assert len(docs["tags"]) == 1

    spec.reset()
    docs = get_docs()
    assert len(docs["tags"]) == 0


def test_custom_specification(app):
    petstore_file = Path(__file__).parent / "static" / "petstore.json"
    petstore_data = petstore_file.read_text()
    app.config.OAS_CUSTOM_FILE = petstore_file

    _, response = app.test_client.get("/docs/openapi.json")
    assert response.text == petstore_data



from __future__ import annotations

from asyncio import sleep
from dataclasses import dataclass
from datetime import datetime, timedelta
from multiprocessing import Manager
from queue import Empty, Full
from signal import SIGINT, SIGTERM
from signal import signal as signal_func
from typing import TYPE_CHECKING, Optional

from sanic.application.constants import ServerStage
from sanic.log import logger


if TYPE_CHECKING:
    from sanic import Sanic


class Stale(ValueError):
    ...


@dataclass
class HealthState:
    name: str
    last: Optional[datetime] = None
    misses: int = 0

    def report(self, timestamp: int) -> None:
        logger.debug(f"Reporting {self.name}")
        if self.misses:
            logger.info(f"Recovered {self.name}")
        self.last = datetime.fromtimestamp(timestamp)
        self.misses = 0

    def missed(self) -> None:
        self.misses += 1
        logger.info(
            f"Missed health check for {self.name} "
            f"({self.misses}/{HealthMonitor.MAX_MISSES})"
        )
        if self.misses >= HealthMonitor.MAX_MISSES:
            raise Stale

    def check(self) -> None:
        if not self.last:
            return

        threshhold = timedelta(
            seconds=(HealthMonitor.MISSED_THRESHHOLD * (self.misses + 1))
        )
        if self.last < (datetime.now() - threshhold):
            self.missed()

    def reset(self) -> None:
        self.misses = 0
        self.last = datetime.now()


def send_healthy(name, queue):
    health = (name, datetime.now().timestamp())
    logger.debug(f"Sending health: {health}", extra={"verbosity": 1})
    try:
        queue.put_nowait(health)
    except Full:
        ...


async def health_check(app: Sanic):
    sent = datetime.now()

    while app.state.stage is ServerStage.SERVING:
        now = datetime.now()
        if sent < now - timedelta(seconds=HealthMonitor.REPORT_INTERVAL):
            send_healthy(app.m.name, app.shared_ctx.health_queue)
            sent = now
        await sleep(0.1)


async def start_health_check(app: Sanic):
    app.add_task(health_check(app), name="health_check")


async def prepare_health_monitor(app, *_):
    HealthMonitor.prepare(app)


async def setup_health_monitor(app, *_):
    health = HealthMonitor(app)
    process_names = [
        process.name for process in app.manager.transient_processes
    ]
    app.manager.manage(
        "HealthMonitor",
        health,
        {
            "process_names": process_names,
            "health_queue": app.shared_ctx.health_queue,
        },
    )


class HealthMonitor:
    MAX_MISSES = 3
    REPORT_INTERVAL = 5
    MISSED_THRESHHOLD = 10

    def __init__(self, app: Sanic):
        self.run = True
        self.monitor_publisher = app.manager.monitor_publisher

    def __call__(self, process_names, health_queue) -> None:
        signal_func(SIGINT, self.stop)
        signal_func(SIGTERM, self.stop)

        now = datetime.now()
        health_state = {
            process_name: HealthState(last=now, name=process_name)
            for process_name in process_names
        }
        while self.run:
            try:
                name, timestamp = health_queue.get(timeout=0.05)
            except Empty:
                ...
            else:
                health_state[name].report(timestamp)

            for state in health_state.values():
                try:
                    state.check()
                except Stale:
                    state.reset()
                    self.monitor_publisher.send(state.name)

    def stop(self, *_):
        self.run = False

    @classmethod
    def prepare(cls, app: Sanic):
        sync_manager = Manager()
        health_queue = sync_manager.Queue(maxsize=app.state.workers * 2)
        app.shared_ctx.health_queue = health_queue

    @classmethod
    def setup(
        cls,
        app: Sanic,
        max_misses: Optional[int] = None,
        report_interval: Optional[int] = None,
        missed_threshhold: Optional[int] = None,
    ):
        HealthMonitor.MAX_MISSES = max_misses or app.config.HEALTH_MAX_MISSES
        HealthMonitor.REPORT_INTERVAL = (
            report_interval or app.config.HEALTH_REPORT_INTERVAL
        )
        HealthMonitor.MISSED_THRESHHOLD = (
            missed_threshhold or app.config.HEALTH_MISSED_THRESHHOLD
        )
        app.main_process_start(prepare_health_monitor)
        app.main_process_ready(setup_health_monitor)
        app.after_server_start(start_health_check)



from dataclasses import asdict, dataclass

from sanic import Sanic, json

from sanic_ext.extras.validation.decorator import validate


@dataclass
class ModelA:
    a: str


@dataclass
class ModelB:
    b: str


def test_both_body_and_query(app: Sanic):
    @app.post("")
    @validate(json=ModelA, query=ModelB)
    async def test(_, body: ModelA, query: ModelB):
        return json(
            {
                "body": asdict(body),
                "query": asdict(query),
            }
        )

    _, response = app.test_client.post("")
    assert response.status == 400

    _, response = app.test_client.post("", params={"b": "bbb"})
    assert response.status == 400

    _, response = app.test_client.post(
        "", params={"b": "bbb"}, json={"a": "aaa"}
    )
    assert response.status == 200
    assert response.json == {
        "body": {"a": "aaa"},
        "query": {"b": "bbb"},
    }



from typing import Any, Dict, Type, get_origin, get_type_hints


def clean_data(model: Type[object], data: Dict[str, Any]) -> Dict[str, Any]:
    hints = get_type_hints(model)
    return {key: _coerce(hints[key], value) for key, value in data.items()}


def _coerce(param_type, value: Any) -> Any:
    if (
        get_origin(param_type) != list
        and isinstance(value, list)
        and len(value) == 1
    ):
        value = value[0]

    return value



import sys
from typing import List, Optional

import pytest

from sanic_ext.utils.typing import contains_annotations, flat_values


class Foo:
    ...


def test_dict_values_nested():
    values = flat_values(
        {
            "level1": {
                "level2": {
                    "level3": {
                        "level4": {"item": 999, "items": [888, 777]},
                        "item": 666,
                    },
                    "item": True,
                },
                "multiple": [
                    {"nested": {"object": None}},
                    {"nested": {"object": None}},
                ],
            },
            "item": True,
        }
    )
    assert values == {999, 888, 777, 666, None, True}


params = [
    ({"foo": str}, True),
    ({"foo": List[str]}, True),
    ({"foo": Optional[str]}, True),
    ({"foo": Foo}, True),
    ({"foo": "str"}, False),
]
if sys.version_info >= (3, 9):
    params.append(({"foo": list[str]}, True))
if sys.version_info >= (3, 10):
    params.append(({"foo": str | None}, True))


@pytest.mark.parametrize("item,expected", params)
def test_contains_annotations(item, expected):
    assert contains_annotations(item) == expected



from sanic import Request, Sanic
from sanic.response import text

from sanic_ext import openapi

from .utils import get_spec


def test_deprecated(app: Sanic):
    @app.route("/test0")
    @openapi.deprecated()
    async def handler0(request: Request):
        return text("ok")

    @app.route("/test1")
    @openapi.deprecated
    async def handler1(request: Request):
        return text("ok")

    @app.route("/test2")
    async def handler2(request: Request):
        """
        openapi:
        ---
        summary: This is a summary.
        deprecated: true
        """
        return text("ok")

    @app.route("/test3")
    @openapi.definition(deprecated=True)
    async def handler3(request: Request):
        return text("ok")

    spec = get_spec(app)
    paths = spec["paths"]
    assert len(paths) == 4
    for i in range(4):
        assert paths[f"/test{i}"]["get"]["deprecated"]



from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Set, Tuple, Type

from sanic.app import Sanic
from sanic.config import Config

from .constructor import Constructor


class InjectionRegistry:
    def __init__(self):
        self._registry: Dict[Type, Optional[Callable[..., Any]]] = {}

    def __getitem__(self, key):
        return self._registry[key]

    def __str__(self) -> str:
        return str(self._registry)

    def __contains__(self, other: Any):
        return other in self._registry

    def get(self, key, default=None):
        return self._registry.get(key, default)

    def register(
        self,
        _type: Type,
        constructor: Optional[Callable[..., Any]],
        request_arg: Optional[str] = None,
    ) -> None:
        constructor = constructor or _type
        constructor = Constructor(constructor, request_arg=request_arg)
        self._registry[_type] = constructor

    def finalize(
        self, app: Sanic, constant_registry: ConstantRegistry, allowed_types
    ):
        for constructor in self._registry.values():
            if isinstance(constructor, Constructor):
                constructor.prepare(
                    app, self, constant_registry, allowed_types
                )

    @property
    def length(self):
        return len(self._registry)


class SignatureRegistry:
    def __init__(self):
        self._registry: Dict[
            str,
            Tuple[
                Dict[str, Tuple[Type, Optional[Callable[..., Any]]]],
                Dict[str, Any],
            ],
        ] = {}

    def __getitem__(self, key):
        return self._registry[key]

    def __str__(self) -> str:
        return str(self._registry)

    def get(self, key, default=None):
        return self._registry.get(key, default)

    def register(
        self,
        route_name: str,
        dependencies: Dict[str, Tuple[Type, Optional[Callable[..., Any]]]],
        constants: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._registry[route_name] = (dependencies, constants or {})


class ConstantRegistry:
    def __init__(self, config: Config):
        self._config = config
        self._registry: Set[str] = set()

    def __str__(self) -> str:
        return str(self._registry)

    def __iter__(self):
        return iter(self._registry)

    def __getitem__(self, key):
        return self._registry[key]

    def __contains__(self, other: Any):
        return other in self._registry

    def register(self, key: str, value: Any, overwrite: bool):
        attribute = key.upper()
        if attribute in self._config and not overwrite:
            raise ValueError(
                f"A value for {attribute} has already been assigned"
            )
        key = key.lower()
        setattr(self._config, attribute, value)
        return self._registry.add(key)

    def get(self, key: str):
        key = key.lower()
        if key not in self._registry:
            raise ValueError
        attribute = key.upper()
        return getattr(self._config, attribute)

    @property
    def length(self):
        return len(self._registry)



import re


def clean_route_name(name: str) -> str:
    parts = name.split(".", 1)
    name = parts[-1]
    for target in ("_", ".", "  "):
        name = name.replace(target, " ")

    return name.title()


def get_uri_filter(app):
    """
    Return a filter function that takes a URI and returns whether it should
    be filter out from the swagger documentation or not.

    Arguments:
        app: The application to take `config.API_URI_FILTER` from. Possible
             values for this config option are: `slash` (to keep URIs that
             end with a `/`), `all` (to keep all URIs). All other values
             default to keep all URIs that don't end with a `/`.

    Returns:
        `True` if the URI should be *filtered out* from the swagger
        documentation, and `False` if it should be kept in the documentation.
    """
    choice = getattr(app.config, "API_URI_FILTER", None)

    if choice == "slash":
        # Keep URIs that end with a /.
        return lambda uri: not uri.endswith("/")

    if choice == "all":
        # Keep all URIs.
        return lambda uri: False

    # Keep URIs that don't end with a /, (special case: "/").
    return lambda uri: len(uri) > 1 and uri.endswith("/")


def remove_nulls(dictionary, deep=True):
    """
    Removes all null values from a dictionary.
    """
    return {
        k: remove_nulls(v, deep) if deep and isinstance(v, dict) else v
        for k, v in dictionary.items()
        if v is not None
    }


def remove_nulls_from_kwargs(**kwargs):
    return remove_nulls(kwargs, deep=False)


def get_blueprinted_routes(app):
    for blueprint in app.blueprints.values():
        if not hasattr(blueprint, "routes"):
            continue

        for route in blueprint.routes:
            if hasattr(route.handler, "view_class"):
                # before sanic 21.3, route.handler could be a number of
                # different things, so have to type check
                for http_method in route.methods:
                    _handler = getattr(
                        route.handler.view_class, http_method.lower(), None
                    )
                    if _handler:
                        yield (blueprint.name, _handler)
            else:
                yield (blueprint.name, route.handler)


def get_all_routes(app, skip_prefix):
    uri_filter = get_uri_filter(app)

    for group in app.router.groups.values():
        uri = f"/{group.path}"

        # prior to sanic 21.3 routes came in both forms
        # (e.g. /test and /test/ )
        # after sanic 21.3 routes come in one form,
        # with an attribute "strict",
        # so we simulate that ourselves:

        uris = [uri]
        if not group.strict and len(uri) > 1:
            alt = uri[:-1] if uri.endswith("/") else f"{uri}/"
            uris.append(alt)

        for uri in uris:
            if uri_filter(uri):
                continue

            if skip_prefix and group.raw_path.startswith(
                skip_prefix.lstrip("/")
            ):
                continue

            for parameter in group.params.values():
                uri = re.sub(
                    f"<{parameter.name}.*?>",
                    f"{{{parameter.name}}}",
                    uri,
                )

            for route in group:
                if getattr(route.extra, "static", False):
                    continue

                method_handlers = [
                    (method, route.handler) for method in route.methods
                ]

                _, name = route.name.split(".", 1)
                yield (
                    uri,
                    name,
                    route.params.values(),
                    method_handlers,
                    route.requirements.get("host"),
                )



import re

from typing import Tuple


# Expression from https://github.com/pypa/packaging
VERSION_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""
PATTERN = re.compile(
    r"^\s*" + VERSION_PATTERN + r"\s*$", re.VERBOSE | re.IGNORECASE
)


def get_version(version) -> Tuple[int, ...]:
    match = PATTERN.search(version)
    if not match:
        raise ValueError(f"Invalid version: {version}")
    return tuple(map(int, match.group("release").split(".")))



import re


CAMEL_TO_SNAKE_PATTERNS = (
    re.compile(r"(.)([A-Z][a-z]+)"),
    re.compile(r"([a-z0-9])([A-Z])"),
)


def camel_to_snake(name: str) -> str:
    for pattern in CAMEL_TO_SNAKE_PATTERNS:
        name = pattern.sub(r"\1_\2", name)
    return name.lower()



import sys
from dataclasses import dataclass, field
from typing import List
from uuid import UUID

import attrs
import pytest
from msgspec import Meta, Struct
from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass as pydataclass

from sanic_ext import openapi

from .utils import get_spec

if sys.version_info >= (3, 9):
    from typing import Annotated


@dataclass
class FooDataclass:
    links: List[UUID]
    priority: int = field(
        metadata={"openapi": {"exclusiveMinimum": 1, "exclusiveMaximum": 10}}
    )
    ident: str = field(
        default="XXXX", metadata={"openapi": {"example": "ABC123"}}
    )


@attrs.define
class FooAttrs:
    links: List[UUID]
    priority: int = attrs.field(
        metadata={"openapi": {"exclusiveMinimum": 1, "exclusiveMaximum": 10}}
    )
    ident: str = attrs.field(
        default="XXXX", metadata={"openapi": {"example": "ABC123"}}
    )


class FooPydanticBaseModel(BaseModel):
    links: List[UUID]
    priority: int = Field(gt=1, lt=10)
    ident: str = Field("XXXX", example="ABC123")


@pydataclass
class FooPydanticDataclass:
    links: List[UUID]
    priority: int = Field(gt=1, lt=10)
    ident: str = Field("XXXX", example="ABC123")


if sys.version_info >= (3, 9):

    class FooStruct(Struct):
        links: List[UUID]
        priority: Annotated[
            int,
            Meta(
                extra={
                    "openapi": {"exclusiveMinimum": 1, "exclusiveMaximum": 10}
                }
            ),
        ]
        ident: Annotated[
            str, Meta(extra={"openapi": {"example": "ABC123"}})
        ] = "XXXX"


models = [
    FooDataclass,
    FooAttrs,
    FooPydanticBaseModel,
    FooPydanticDataclass,
]

if sys.version_info >= (3, 9):
    models.append(FooStruct)


@pytest.mark.parametrize("Foo", models)
def test_models(app, Foo):
    @app.get("/")
    @openapi.definition(body={"application/json": Foo})
    async def handler(_):
        ...

    spec = get_spec(app)
    foo_props = spec["paths"]["/"]["get"]["requestBody"]["content"][
        "application/json"
    ]["schema"]["properties"]

    assert foo_props["links"] == {
        "title": "Links",
        "type": "array",
        "items": {"type": "string", "format": "uuid"},
    }
    assert foo_props["ident"] == {
        "title": "Ident",
        "type": "string",
        "default": "XXXX",
        "example": "ABC123",
    }
    assert foo_props["priority"] == {
        "title": "Priority",
        "type": "integer",
        "format": "int32",
        "exclusiveMinimum": 1,
        "exclusiveMaximum": 10,
    }



from unittest.mock import Mock

import pytest
from sanic import Sanic
from sanic.compat import Header
from sanic.response import json
from sanic_testing.reusable import ReusableClient

from sanic_ext import CountedRequest


@pytest.fixture(autouse=True)
def reset_counter():
    yield
    CountedRequest.reset_count()


def test_counter_increments(app: Sanic):
    app.request_class = CountedRequest

    @app.get("/")
    async def handler(request: CountedRequest):
        return json({"count": request.count})

    @app.get("/info")
    async def info(request: CountedRequest):
        return json({"state": request.app.m.state})

    with ReusableClient(app) as client:
        for i in range(1, 10):
            _, response = client.get("/")
            assert response.json["count"] == i


def test_counter_increment_on_state(app: Sanic):
    mock = Mock()
    mock.state = {}
    app.multiplexer = mock

    for i in range(1, 10):
        CountedRequest(b"/", Header({}), "", "", Mock(), app)
        assert CountedRequest.count == i



from ..base import Extension
from .injector import add_injection
from .registry import ConstantRegistry, InjectionRegistry


class InjectionExtension(Extension):
    name = "injection"

    def startup(self, bootstrap) -> None:
        self.constant_registry = ConstantRegistry(self.app.config)
        self.registry = InjectionRegistry()
        add_injection(self.app, self.registry, self.constant_registry)
        bootstrap._injection_registry = self.registry
        bootstrap._constant_registry = self.constant_registry
        if self.config.INJECTION_LOAD_CUSTOM_CONSTANTS:
            self.app.ext.load_constants()

    def label(self):
        return (
            f"{self.registry.length} dependencies; "
            f"{self.constant_registry.length} constants"
        )



import pytest
from sanic import Sanic
from sanic.response import empty, text

from sanic_ext.bootstrap import Extend


def test_trace_and_connect_available(app: Sanic):
    @app.route("/", methods=["trace", "connect"])
    async def handler(_):
        return empty()

    _, response = app.test_client.request("", http_method="trace")
    assert response.status == 204
    _, response = app.test_client.request("", http_method="connect")
    assert response.status == 204
    _, response = app.test_client.request("", http_method="get")
    assert response.status == 405


def test_auto_head(app: Sanic, get_docs):
    app.config.TOUCHUP = False

    @app.get("/foo")
    async def foo_handler(_):
        return text("...")

    assert app.config.HTTP_AUTO_HEAD
    _, response = app.test_client.head("/foo")
    assert response.status == 200
    assert len(response.body) == 0
    assert int(response.headers["content-length"]) == 3

    schema = get_docs()
    assert "get" in schema["paths"]["/foo"]


def test_auto_options(app: Sanic, get_docs):
    @app.post("/foo")
    async def foo_handler(_):
        return text("...")

    _, response = app.test_client.options("/foo")
    assert response.status == 204
    assert len(response.body) == 0
    assert "POST" in response.headers["allow"]
    assert "OPTIONS" in response.headers["allow"]

    schema = get_docs()
    assert "post" in schema["paths"]["/foo"]


def test_auto_trace(bare_app: Sanic):
    Extend(bare_app, config={"http_auto_trace": True})

    @bare_app.get("/foo")
    async def foo_handler(_):
        return text("...")

    request, response = bare_app.test_client.request(
        "/foo", http_method="trace"
    )
    assert response.status == 200
    assert response.body.startswith(request.head)


def test_auto_head_with_vhosts(app: Sanic, get_docs):
    @app.get("/foo", host="one.com", name="one")
    async def foo_handler_one(_):
        return text(".")

    @app.get("/foo", host="two.com", name="two")
    async def foo_handler_two(_):
        return text("..")

    assert app.config.HTTP_AUTO_HEAD
    _, response = app.test_client.head("/foo", headers={"host": "one.com"})
    assert response.status == 200
    assert len(response.body) == 0
    assert int(response.headers["content-length"]) == 1

    _, response = app.test_client.head("/foo", headers={"host": "two.com"})
    assert response.status == 200
    assert len(response.body) == 0
    assert int(response.headers["content-length"]) == 2

    schema = get_docs()
    assert "get" in schema["paths"]["/foo"]


def test_auto_options_with_vhosts(app: Sanic, get_docs):
    @app.post("/foo", host="one.com", name="one")
    async def foo_handler_one(_):
        return text(".")

    @app.post("/foo", host="two.com", name="two")
    async def foo_handler_two(_):
        return text("..")

    assert app.config.HTTP_AUTO_OPTIONS
    _, response = app.test_client.options("/foo", headers={"host": "one.com"})
    assert response.status == 204
    assert len(response.body) == 0
    assert "POST" in response.headers["allow"]
    assert "OPTIONS" in response.headers["allow"]

    _, response = app.test_client.options("/foo", headers={"host": "two.com"})
    assert response.status == 204
    assert len(response.body) == 0
    assert "POST" in response.headers["allow"]
    assert "OPTIONS" in response.headers["allow"]


# This test also appears in Core Sanic tests but is added here as well
# because of: https://github.com/sanic-org/sanic-ext/issues/148
@pytest.mark.parametrize("unquote", [True, False, None])
def test_unquote_add_route(app, unquote):
    async def handler1(_, foo):
        return text(foo)

    app.add_route(handler1, "/<foo>", unquote=unquote)
    value = "啊" if unquote else r"%E5%95%8A"

    _, response = app.test_client.get("/啊")
    assert response.text == value

    _, response = app.test_client.get(r"/%E5%95%8A")
    assert response.text == value



import inspect
import warnings

import yaml


class OpenAPIDocstringParser:
    def __init__(self, docstring: str):
        """
        Args:
            docstring (str): docstring of function to be parsed
        """
        if docstring is None:
            docstring = ""
        self.docstring = inspect.cleandoc(docstring)

    def to_openAPI_2(self) -> dict:
        """
        Returns:
            json style dict: dict to be read for the path by swagger 2.0 UI
        """
        raise NotImplementedError()

    def to_openAPI_3(self) -> dict:
        """
        Returns:
            json style dict: dict to be read for the path by swagger 3.0.0 UI
        """
        raise NotImplementedError()


class YamlStyleParametersParser(OpenAPIDocstringParser):
    def _parse_no_yaml(self, doc: str) -> dict:
        """
        Args:
            doc (str): section of doc before yaml, or full section of doc
        Returns:
            json style dict: dict to be read for the path by swagger UI
        """
        # clean again in case further indentation can be removed,
        # usually this do nothing...
        doc = inspect.cleandoc(doc)

        if len(doc) == 0:
            return {}

        lines = doc.split("\n")

        if len(lines) == 1:
            return {"summary": lines[0]}
        else:
            summary = lines.pop(0)

            # remove empty lines at the beginning of the description
            while len(lines) and lines[0].strip() == "":
                lines.pop(0)

            if len(lines) == 0:
                return {"summary": summary}
            else:
                # use html tag to preserve linebreaks
                return {"summary": summary, "description": "<br>".join(lines)}

    def _parse_yaml(self, doc: str) -> dict:
        """
        Args:
            doc (str): section of doc detected as openapi yaml
        Returns:
            json style dict: dict to be read for the path by swagger UI
        Warns:
            UserWarning if the yaml couldn't be parsed
        """
        try:
            return yaml.safe_load(doc)
        except Exception as e:
            warnings.warn(
                "error parsing openAPI yaml, ignoring it. ({})".format(e)
            )
            return {}

    def _parse_all(self) -> dict:
        if "openapi:\n" not in self.docstring:
            return self._parse_no_yaml(self.docstring)

        predoc, yamldoc = self.docstring.split("openapi:\n", 1)

        conf = self._parse_no_yaml(predoc)
        conf.update(self._parse_yaml(yamldoc))
        return conf

    def to_openAPI_2(self) -> dict:
        return self._parse_all()

    def to_openAPI_3(self) -> dict:
        return self._parse_all()



from typing import List

import pydantic
from pydantic.dataclasses import dataclass
from sanic import json
from sanic.views import HTTPMethodView

from sanic_ext import validate
from sanic_ext.exceptions import ValidationError

SNOOPY_DATA = {"name": "Snoopy", "alter_ego": ["Flying Ace", "Joe Cool"]}


def test_validate_json(app):
    @dataclass
    class Pet:
        name: str
        alter_ego: List[str]

    @app.post("/function")
    @validate(json=Pet)
    async def handler(_, body: Pet):
        return json(
            {
                "is_pet": isinstance(body, Pet),
                "pet": {"name": body.name, "alter_ego": body.alter_ego},
            }
        )

    class MethodView(HTTPMethodView, attach=app, uri="/method"):
        decorators = [validate(json=Pet)]

        async def post(self, _, body: Pet):
            return json(
                {
                    "is_pet": isinstance(body, Pet),
                    "pet": {"name": body.name, "alter_ego": body.alter_ego},
                }
            )

    _, response = app.test_client.post("/function", json=SNOOPY_DATA)
    assert response.status == 200
    assert response.json["is_pet"]
    assert response.json["pet"] == SNOOPY_DATA

    _, response = app.test_client.post("/method", json=SNOOPY_DATA)
    assert response.status == 200
    assert response.json["is_pet"]
    assert response.json["pet"] == SNOOPY_DATA


def test_validate_form(app):
    @dataclass
    class Pet:
        name: str
        alter_ego: List[str]

    @app.post("/function")
    @validate(form=Pet)
    async def handler(_, body: Pet):
        return json(
            {
                "is_pet": isinstance(body, Pet),
                "pet": {"name": body.name, "alter_ego": body.alter_ego},
            }
        )

    class MethodView(HTTPMethodView, attach=app, uri="/method"):
        decorators = [validate(form=Pet)]

        async def post(self, _, body: Pet):
            return json(
                {
                    "is_pet": isinstance(body, Pet),
                    "pet": {"name": body.name, "alter_ego": body.alter_ego},
                }
            )

    _, response = app.test_client.post("/function", data=SNOOPY_DATA)
    assert response.status == 200
    assert response.json["is_pet"]
    assert response.json["pet"] == SNOOPY_DATA

    _, response = app.test_client.post("/method", data=SNOOPY_DATA)
    assert response.status == 200
    assert response.json["is_pet"]
    assert response.json["pet"] == SNOOPY_DATA


def test_validate_query(app):
    @dataclass
    class Search:
        q: str

    @app.get("/function")
    @validate(query=Search)
    async def handler(_, query: Search):
        return json({"q": query.q, "is_search": isinstance(query, Search)})

    class MethodView(HTTPMethodView, attach=app, uri="/method"):
        decorators = [validate(query=Search)]

        async def get(self, _, query: Search):
            return json({"q": query.q, "is_search": isinstance(query, Search)})

    _, response = app.test_client.get("/function", params={"q": "Snoopy"})
    assert response.status == 200
    assert response.json["is_search"]
    assert response.json["q"] == "Snoopy"

    _, response = app.test_client.get("/method", params={"q": "Snoopy"})
    assert response.status == 200
    assert response.json["is_search"]
    assert response.json["q"] == "Snoopy"


class User(pydantic.BaseModel):
    name: str
    age: int
    email: str


def test_success_validate_form_custom_message(app):
    @app.post("/user")
    @validate(form=User)
    async def create_user(request, body: User):
        return json(body.dict())

    user = {"name": "Alison", "age": 25, "email": "alison@almeida.com"}
    _, response = app.test_client.post("/user", data=user)
    assert response.status == 200


def test_error_validate_form_custom_message(app):
    async def server_error_validate_form(request, exception: ValidationError):
        error = exception.extra["exception"]
        assert isinstance(error, pydantic.ValidationError)
        return json(error.json(), status=400)

    @app.post("/user")
    @validate(form=User)
    async def create_user(request, body: User):
        return json(body.dict())

    user = {"name": "Alison", "age": 25}

    app.error_handler.add(ValidationError, server_error_validate_form)
    _, response = app.test_client.post("/user", data=user)
    assert response.status == 400



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from itertools import count
from typing import Optional
from uuid import UUID

import pytest
from sanic import Request, json, text
from sanic.exceptions import SanicException
from sanic.views import HTTPMethodView

from sanic_ext import Extend


@dataclass
class Name:
    name: str


@dataclass
class PersonID:
    person_id: int


@dataclass
class NamedPerson:
    name: str


@dataclass
class Person:
    person_id: PersonID
    name: str
    age: int

    @classmethod
    async def create(cls, person_id: int) -> Person:
        return cls(person_id=PersonID(person_id), name="noname", age=111)


@dataclass
class AsyncName:
    name: str

    def __await__(self):
        return self


counter = count()


class A:
    @classmethod
    def create(cls, request: Request):
        next(counter)
        return cls()


class B:
    def __init__(self, a: A):
        self.a = a

    @classmethod
    def create(cls, request: Request, a: A):
        next(counter)
        return cls(a)


class C:
    def __init__(self, b: B):
        self.b = b

    @classmethod
    def create(cls, request: Request, b: B):
        next(counter)
        return cls(b)


class D:
    def __init__(self, a: A):
        self.a = a

    @classmethod
    def create(cls, request: Request, a: A):
        next(counter)
        return cls(a)


class E:
    def __init__(self, c: C, d: D):
        self.c = c
        self.d = d

    @classmethod
    def create(cls, request: Request, c: C, d: D):
        next(counter)
        return cls(c, d)


class Alpha:
    ...


class Beta:
    def __init__(self, alpha: Alpha) -> None:
        self.alpha = alpha


class Gamma:
    def __init__(self, beta: Beta, request: Optional[Request] = None) -> None:
        self.beta = beta
        self.request = request


class Delta:
    def __init__(self, gamma: Gamma, request: Request) -> None:
        self.gamma = gamma
        self.request = request


@dataclass
class Foo:
    ident: UUID


def make_gamma_with_request(beta: Beta, request: Request):
    return Gamma(beta, request)


def make_gamma_without_request(beta: Beta):
    return Gamma(beta)


def test_injection_not_allowed_when_ext_disabled(bare_app):
    ext = Extend(bare_app, built_in_extensions=False)

    with pytest.raises(
        SanicException, match="Injection extension not enabled"
    ):
        ext.add_dependency(1, 2)


def test_injection_of_matched_object(app):
    @app.get("/person/<name:str>")
    def handler(request, name: Name):
        request.ctx.name = name
        return text(name.name)

    app.ext.add_dependency(Name)

    request, response = app.test_client.get("/person/george")

    assert response.body == b"george"
    assert isinstance(request.ctx.name, Name)
    assert request.ctx.name.name == "george"


def test_injection_of_matched_object_as_deprecated_injection(app):
    @app.get("/person/<name:str>")
    def handler(request, name: Name):
        request.ctx.name = name
        return text(name.name)

    message = (
        "The 'ext.injection' method has been deprecated and will be removed "
        "in v22.6. Please use 'ext.add_dependency' instead."
    )
    with pytest.warns(DeprecationWarning, match=message):
        app.ext.injection(Name)

    request, response = app.test_client.get("/person/george")

    assert response.body == b"george"
    assert isinstance(request.ctx.name, Name)
    assert request.ctx.name.name == "george"


def test_injection_of_simple_object(app):
    @app.get("/person/<name>")
    def handler(request, name: str, person: NamedPerson):
        request.ctx.person = person
        return text(person.name)

    app.ext.add_dependency(NamedPerson)

    request, response = app.test_client.get("/person/george")

    assert response.body == b"george"
    assert isinstance(request.ctx.person, NamedPerson)
    assert request.ctx.person.name == "george"


def test_injection_of_object_with_constructor(app):
    @app.get("/person/<person_id:int>")
    async def person_details(request, person_id: PersonID, person: Person):
        request.ctx.person_id = person_id
        request.ctx.person = person
        return text(
            f"{person.person_id.person_id}\n{person.name}\n{person.age}"
        )

    app.ext.add_dependency(Person, Person.create)
    app.ext.add_dependency(PersonID)

    request, response = app.test_client.get("/person/999")

    assert response.body == b"999\nnoname\n111"
    assert isinstance(request.ctx.person_id, PersonID)
    assert isinstance(request.ctx.person, Person)
    assert request.ctx.person.person_id == request.ctx.person_id
    assert request.ctx.person.person_id.person_id == 999
    assert request.ctx.person.name == "noname"
    assert request.ctx.person.age == 111


def test_injection_on_cbv(app):
    class View(HTTPMethodView, attach=app, uri="/person/<name:str>"):
        async def get(self, request, name: Name):
            request.ctx.name = name
            return text(name.name)

        @staticmethod
        async def post(request, name: Name):
            request.ctx.name = name
            return text(name.name)

    app.ext.add_dependency(Name)

    for client in (app.test_client.get, app.test_client.post):
        request, response = client("/person/george")

        assert response.body == b"george"
        assert isinstance(request.ctx.name, Name)
        assert request.ctx.name.name == "george"


def test_nested_dependencies(app):
    app.ext.add_dependency(A, A.create)
    app.ext.add_dependency(B, B.create)
    app.ext.add_dependency(C, C.create)
    app.ext.add_dependency(D, D.create)
    app.ext.add_dependency(E, E.create)

    @app.get("/")
    async def nested(request: Request, c: C, e: E):
        return json(
            [
                isinstance(c, C),
                isinstance(c.b, B),
                isinstance(c.b.a, A),
                isinstance(e, E),
                isinstance(e.c, C),
                isinstance(e.d, D),
                isinstance(e.d.a, A),
            ]
        )

    _, response = app.test_client.get("/")

    assert all(response.json)
    # TODO:
    # - After implementing https://github.com/sanic-org/sanic-ext/issues/77
    #   this should be == 5
    assert next(counter) == 9


def test_injection_on_websocket(app):
    ev = asyncio.Event()

    app.ext.dependency(A())

    @app.websocket("/foo")
    async def handler(request, ws, foo: A):
        assert isinstance(foo, A)
        if isinstance(foo, A):
            ev.set()

    request, response = app.test_client.websocket("/foo")
    assert ev.is_set()


def test_injection_of_awaitable_variable_in_do_cast(app):
    """Test for do_cast() iscoroutine() check"""

    @app.get("/person/<name:str>")
    def handler(request, name: AsyncName):
        request.ctx.name = name
        return text(name.name)

    app.ext.add_dependency(AsyncName)

    request, response = app.test_client.get("/person/george")

    assert response.body == b"george"
    assert isinstance(request.ctx.name, AsyncName)
    assert request.ctx.name.name == "george"


def test_injection_of_awaitable_variable_in_call(app):
    """Test for __call__() iscoroutine() check"""

    @app.get("/person/<name:str>")
    def handler(request, name: AsyncName):
        request.ctx.name = name
        return text(name.name)

    def test():
        return AsyncName("george")

    app.ext.dependency(test())

    request, response = app.test_client.get("/person/george")

    assert response.body == b"george"
    assert isinstance(request.ctx.name, AsyncName)
    assert request.ctx.name.name == "george"


def test_injection_class_constructors(app):
    app.ext.add_dependency(Alpha)
    app.ext.add_dependency(Beta)

    @app.get("/")
    def handler(request: Request, beta: Beta):
        return json({"is_beta": isinstance(beta, Beta)})

    _, response = app.test_client.get("/")
    assert response.json == {"is_beta": True}


def test_injection_class_constructors_with_optional_request(app):
    app.ext.add_dependency(Alpha)
    app.ext.add_dependency(Beta)
    app.ext.add_dependency(Gamma)

    @app.get("/")
    def handler(request: Request, gamma: Gamma):
        return json(
            {
                "is_gamma": isinstance(gamma, Gamma),
                "has_request": bool(gamma.request),
            }
        )

    _, response = app.test_client.get("/")
    assert response.json == {"is_gamma": True, "has_request": True}


def test_injection_class_constructors_with_request(app):
    app.ext.add_dependency(Alpha)
    app.ext.add_dependency(Beta)
    app.ext.add_dependency(Gamma)
    app.ext.add_dependency(Delta)

    @app.get("/")
    def handler(request: Request, delta: Delta):
        return json(
            {
                "is_delta": isinstance(delta, Delta),
                "has_request": bool(delta.request),
            }
        )

    _, response = app.test_client.get("/")
    assert response.json == {"is_delta": True, "has_request": True}


def test_injection_class_constructors_with_func_and_request(app):
    app.ext.add_dependency(Alpha)
    app.ext.add_dependency(Beta)
    app.ext.add_dependency(Gamma, make_gamma_with_request)

    @app.get("/")
    def handler(request: Request, gamma: Gamma):
        return json(
            {
                "is_gamma": isinstance(gamma, Gamma),
                "has_request": bool(gamma.request),
            }
        )

    _, response = app.test_client.get("/")
    assert response.json == {"is_gamma": True, "has_request": True}


def test_injection_class_constructors_with_func_and_no_request(app):
    app.ext.add_dependency(Alpha)
    app.ext.add_dependency(Beta)
    app.ext.add_dependency(Gamma, make_gamma_without_request)

    @app.get("/")
    def handler(request: Request, gamma: Gamma):
        return json(
            {
                "is_gamma": isinstance(gamma, Gamma),
                "has_request": bool(gamma.request),
            }
        )

    _, response = app.test_client.get("/")
    assert response.json == {"is_gamma": True, "has_request": False}


def test_injection_of_lambda_properties(app):
    @app.get("/foo")
    def handler(request, foo: Foo):
        return json(request.id == foo.ident and isinstance(foo.ident, UUID))

    app.ext.add_dependency(Foo, lambda request: Foo(request.id), "request")

    _, response = app.test_client.get("/foo")

    assert response.body == b"true"



import pytest
from sanic import Sanic

from .utils import get_spec


@pytest.fixture(autouse=True)
def handlers(app: Sanic):
    @app.route("/test1")
    async def handler1(_):
        ...

    @app.route("/test2/1")
    async def handler2(_):
        ...

    @app.route("/test3/")
    async def handler3(_):
        ...

    @app.route("/test4/2/")
    async def handler4(_):
        ...


def test_paths_with_all_uri_filter(app: Sanic):
    app.config.API_URI_FILTER = "all"

    spec = get_spec(app)
    assert list(spec["paths"].keys()) == [
        "/test1",
        "/test1/",
        "/test2/1",
        "/test2/1/",
        "/test3",
        "/test3/",
        "/test4/2",
        "/test4/2/",
    ]


def test_paths_with_slash_uri_filter(app: Sanic):
    app.config.API_URI_FILTER = "slash"
    spec = get_spec(app)

    assert list(spec["paths"].keys()) == [
        "/test1/",
        "/test2/1/",
        "/test3/",
        "/test4/2/",
    ]


def test_paths_without_uri_filter(app: Sanic):
    spec = get_spec(app)
    assert list(spec["paths"].keys()) == [
        "/test1",
        "/test2/1",
        "/test3",
        "/test4/2",
    ]



from __future__ import annotations

from functools import partial
from inspect import getmembers, isclass, isfunction
from typing import Any, Callable, Dict, Optional, Tuple, Type, get_type_hints

from sanic import Sanic
from sanic.constants import HTTP_METHODS

from sanic_ext.config import PRIORITY
from sanic_ext.extensions.injection.constructor import gather_args

from .registry import ConstantRegistry, InjectionRegistry, SignatureRegistry


def add_injection(
    app: Sanic,
    injection_registry: InjectionRegistry,
    constant_registry: ConstantRegistry,
) -> None:
    signature_registry = _setup_signature_registry(
        app, injection_registry, constant_registry
    )

    @app.listener("before_server_start", priority=PRIORITY)
    async def finalize_injections(app: Sanic, _):
        router_converters = set(
            allowed[0] for allowed in app.router.regex_types.values()
        )
        router_types = set()
        for converter in router_converters:
            if isclass(converter):
                router_types.add(converter)
            elif isfunction(converter):
                hints = get_type_hints(converter)
                if return_type := hints.get("return"):
                    router_types.add(return_type)
        injection_registry.finalize(app, constant_registry, router_types)

    injection_signal = app.ext.config.INJECTION_SIGNAL
    injection_priority = app.ext.config.INJECTION_PRIORITY

    @app.signal(injection_signal, priority=injection_priority)
    async def inject_kwargs(request, **_):
        nonlocal signature_registry

        for name in (
            request.route.name,
            f"{request.route.name}_{request.method.lower()}",
        ):
            dependencies, constants = signature_registry.get(
                name, (None, None)
            )
            if dependencies or constants:
                break

        if dependencies:
            injected_args = await gather_args(
                dependencies, request, **request.match_info
            )
            request.match_info.update(injected_args)
        if constants:
            request.match_info.update(constants)


def _http_method_predicate(member):
    return isfunction(member) and member.__name__ in HTTP_METHODS


def _setup_signature_registry(
    app: Sanic,
    injection_registry: InjectionRegistry,
    constant_registry: ConstantRegistry,
) -> SignatureRegistry:
    registry = SignatureRegistry()

    @app.listener("before_server_start", priority=PRIORITY - 1)
    async def setup_signatures(app, _):
        nonlocal registry

        for route in app.router.routes:
            if ".openapi." in route.name:
                continue
            handlers = [(route.name, route.handler)]
            viewclass = getattr(route.handler, "view_class", None)
            if viewclass:
                handlers = [
                    (f"{route.name}_{name}", member)
                    for name, member in getmembers(
                        viewclass, _http_method_predicate
                    )
                ]
            for name, handler in handlers:
                if route_handler := getattr(
                    handler, "__route_handler__", None
                ):
                    handler = route_handler
                if isinstance(handler, partial):
                    if handler.func == app._websocket_handler:
                        handler = handler.args[0]
                    else:
                        handler = handler.func
                try:
                    hints = get_type_hints(handler)
                except TypeError:
                    continue

                dependencies: Dict[
                    str, Tuple[Type, Optional[Callable[..., Any]]]
                ] = {}
                constants: Dict[str, Any] = {}
                for param, annotation in hints.items():
                    if annotation in injection_registry:
                        dependencies[param] = (
                            annotation,
                            injection_registry[annotation],
                        )
                    if param in constant_registry:
                        constants[param] = app.config[param.upper()]

                registry.register(name, dependencies, constants)

    return registry



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Optional, Sequence, Union, cast

from sanic_ext.extensions.openapi.constants import (
    SecuritySchemeAuthorization,
    SecuritySchemeLocation,
    SecuritySchemeType,
)

from ...utils.route import remove_nulls, remove_nulls_from_kwargs
from .autodoc import YamlStyleParametersParser
from .definitions import (
    Any,
    Components,
    Contact,
    Dict,
    ExternalDocumentation,
    Flows,
    Info,
    License,
    List,
    OpenAPI,
    Operation,
    Parameter,
    PathItem,
    RequestBody,
    Response,
    SecurityRequirement,
    SecurityScheme,
    Server,
    Tag,
)


if TYPE_CHECKING:
    from sanic import Sanic


class OperationBuilder:
    summary: str
    description: str
    operationId: str
    requestBody: RequestBody
    externalDocs: ExternalDocumentation
    tags: List[str]
    security: List[Any]
    parameters: List[Parameter]
    responses: Dict[str, Response]
    callbacks: List[str]  # TODO
    deprecated: bool = False

    def __init__(self):
        self.tags = []
        self.security = []
        self.parameters = []
        self.responses = {}
        self._default = {}
        self._autodoc = None
        self._exclude = False
        self._allow_autodoc = True
        self._app: Optional[Sanic] = None

    def name(self, value: str):
        self.operationId = value

    def describe(self, summary: str = None, description: str = None):
        if summary:
            self.summary = summary

        if description:
            self.description = description

    def document(self, url: str, description: str = None):
        self.externalDocs = ExternalDocumentation.make(url, description)

    def tag(self, *args: str):
        for arg in args:
            if isinstance(arg, Tag):
                arg = arg.fields["name"]
            self.tags.append(arg)

    def deprecate(self):
        self.deprecated = True

    def body(self, content: Any, **kwargs):
        self.requestBody = RequestBody.make(content, **kwargs)

    def parameter(
        self, name: str, schema: Any, location: str = "query", **kwargs
    ):
        self.parameters.append(
            Parameter.make(name, schema, location, **kwargs)
        )

    def response(
        self, status, content: Any = None, description: str = None, **kwargs
    ):
        response = Response.make(content, description, **kwargs)
        if status in self.responses:
            self.responses[status]._fields["content"].update(
                response.fields["content"]
            )
        else:
            self.responses[status] = response

    def secured(self, *args, **kwargs):
        if not kwargs and len(args) == 1 and isinstance(args[0], dict):
            items = args[0]
        else:
            items = {**{v: [] for v in args}, **kwargs}
        gates = {}

        for name, params in items.items():
            gate = name.__name__ if isinstance(name, type) else name
            gates[gate] = params

        self.security.append(gates)

    def disable_autodoc(self):
        self._allow_autodoc = False

    def build(self):
        operation_dict = self._build_merged_dict()
        if "responses" not in operation_dict:
            # todo -- look into more consistent default response format
            operation_dict["responses"] = {"default": {"description": "OK"}}

        return Operation(**operation_dict)

    def _build_merged_dict(self):
        defined_dict = self.__dict__.copy()
        autodoc_dict = self._autodoc or {}
        default_dict = self._default
        merged_dict = {}

        for d in (default_dict, autodoc_dict, defined_dict):
            cleaned = {
                k: v for k, v in d.items() if v and not k.startswith("_")
            }
            merged_dict.update(cleaned)

        return merged_dict

    def autodoc(self, docstring: str):
        y = YamlStyleParametersParser(docstring)
        self._autodoc = y.to_openAPI_3()

    def exclude(self, flag: bool = True):
        self._exclude = flag


class OperationStore(defaultdict):
    _singleton = None

    def __new__(cls) -> Any:
        if not cls._singleton:
            cls._singleton = super().__new__(cls)
        return cls._singleton

    def __init__(self):
        super().__init__(OperationBuilder)

    @classmethod
    def reset(cls):
        cls._singleton = None


class SpecificationBuilder:
    _urls: List[str]
    _title: str
    _version: str
    _description: Optional[str]
    _terms: Optional[str]
    _contact: Contact
    _license: License
    _paths: Dict[str, Dict[str, OperationBuilder]]
    _tags: Dict[str, Tag]
    _security: List[SecurityRequirement]
    _components: Dict[str, Any]
    _servers: List[Server]
    # _components: ComponentsBuilder
    # deliberately not included
    _singleton: Optional[SpecificationBuilder] = None

    def __new__(cls) -> SpecificationBuilder:
        if not cls._singleton:
            cls._singleton = super().__new__(cls)
            cls._setup_instance(cls._singleton)
        return cast(SpecificationBuilder, cls._singleton)

    @classmethod
    def _setup_instance(cls, instance):
        instance._components = defaultdict(dict)
        instance._contact = None
        instance._description = None
        instance._external = None
        instance._license = None
        instance._paths = defaultdict(dict)
        instance._servers = []
        instance._tags = {}
        instance._security = []
        instance._terms = None
        instance._title = None
        instance._urls = []
        instance._version = None

    @classmethod
    def reset(cls):
        cls._singleton = None

    @property
    def tags(self):
        return self._tags

    @property
    def security(self):
        return self._security

    def url(self, value: str):
        self._urls.append(value)

    def describe(
        self,
        title: str,
        version: str,
        description: Optional[str] = None,
        terms: Optional[str] = None,
    ):
        self._title = title
        self._version = version
        self._description = description
        self._terms = terms

    def _do_describe(
        self,
        title: str,
        version: str,
        description: Optional[str] = None,
        terms: Optional[str] = None,
    ):
        if any([self._title, self._version, self._description, self._terms]):
            return
        self.describe(title, version, description, terms)

    def tag(self, name: str, description: Optional[str] = None, **kwargs):
        self._tags[name] = Tag(name, description=description, **kwargs)

    def external(self, url: str, description: Optional[str] = None, **kwargs):
        self._external = ExternalDocumentation(url, description=description)

    def secured(
        self,
        name: str = None,
        value: Optional[Union[str, Sequence[str]]] = None,
    ):
        if value is None:
            value = []
        elif isinstance(value, str):
            value = [value]
        else:
            value = list(value)
        self._security.append(SecurityRequirement(name=name, value=value))

    def contact(self, name: str = None, url: str = None, email: str = None):
        kwargs = remove_nulls_from_kwargs(name=name, url=url, email=email)
        self._contact = Contact(**kwargs)

    def _do_contact(
        self, name: str = None, url: str = None, email: str = None
    ):
        if self._contact:
            return

        self.contact(name, url, email)

    def license(self, name: str = None, url: str = None):
        if name is not None:
            self._license = License(name, url=url)

    def _do_license(self, name: str = None, url: str = None):
        if self._license:
            return

        self.license(name, url)

    def operation(self, path: str, method: str, operation: OperationBuilder):
        for _tag in operation.tags:
            if _tag in self._tags.keys():
                continue

            self._tags[_tag] = Tag(_tag)

        self._paths[path][method.lower()] = operation

    def add_component(self, location: str, name: str, obj: Any):
        self._components[location].update({name: obj})

    def has_component(self, location: str, name: str) -> bool:
        return name in self._components.get(location, {})

    def add_security_scheme(
        self,
        ident: str,
        type: Union[str, SecuritySchemeType],
        *,
        bearer_format: Optional[str] = None,
        description: Optional[str] = None,
        flows: Optional[Union[Flows, Dict[str, Any]]] = None,
        location: Union[
            str, SecuritySchemeLocation
        ] = SecuritySchemeLocation.HEADER,
        name: str = "authorization",
        openid_connect_url: Optional[str] = None,
        scheme: Union[
            str, SecuritySchemeAuthorization
        ] = SecuritySchemeAuthorization.BEARER,
    ):
        if isinstance(type, str):
            type = SecuritySchemeType(type)
        if isinstance(location, str):
            location = SecuritySchemeLocation(location)

        kwargs: Dict[str, Any] = {"type": type, "description": description}

        if type is SecuritySchemeType.API_KEY:
            kwargs["location"] = location
            kwargs["name"] = name
        elif type is SecuritySchemeType.HTTP:
            kwargs["scheme"] = scheme
            kwargs["bearerFormat"] = bearer_format
        elif type is SecuritySchemeType.OAUTH2:
            kwargs["flows"] = flows
        elif type is SecuritySchemeType.OPEN_ID_CONNECT:
            kwargs["openIdConnectUrl"] = openid_connect_url

        self.add_component(
            "securitySchemes",
            ident,
            SecurityScheme(**kwargs),
        )  # type: ignore

    def raw(self, data):
        if "info" in data:
            self.describe(
                data["info"].get("title"),
                data["info"].get("version"),
                data["info"].get("description"),
                data["info"].get("terms"),
            )

        if "servers" in data:
            for server in data["servers"]:
                self._servers.append(Server(**server))

        if "paths" in data:
            self._paths.update(data["paths"])

        if "components" in data:
            for location, component in data["components"].items():
                self._components[location].update(component)

        if "security" in data:
            for security in data["security"]:
                if not security:
                    self.secured()
                else:
                    for key, value in security.items():
                        self.secured(key, value)

        if "tags" in data:
            for tag in data["tags"]:
                self.tag(**tag)

        if "externalDocs" in data:
            self.external(**data["externalDocs"])

    def build(self, app: Sanic) -> OpenAPI:
        info = self._build_info()
        paths = self._build_paths(app)
        tags = self._build_tags()
        security = self._build_security()

        url_servers = getattr(self, "_urls", None)
        servers = self._servers
        existing = [
            server.fields["url"].strip("/") for server in self._servers
        ]
        if url_servers is not None:
            for url_server in url_servers:
                if url_server.strip("/") not in existing:
                    servers.append(Server(url=url_server))

        components = (
            Components(**self._components) if self._components else None
        )

        return OpenAPI(
            info,
            paths,
            tags=tags,
            servers=servers,
            security=security,
            components=components,
            externalDocs=self._external,
        )

    def _build_info(self) -> Info:
        kwargs = remove_nulls(
            {
                "description": self._description,
                "termsOfService": self._terms,
                "license": self._license,
                "contact": self._contact,
            },
            deep=False,
        )

        return Info(self._title, self._version, **kwargs)

    def _build_tags(self):
        return [self._tags[k] for k in self._tags]

    def _build_paths(self, app: Sanic) -> Dict:
        paths = {}

        for path, operations in self._paths.items():
            paths[path] = PathItem(
                **{
                    k: v if isinstance(v, dict) else v.build()
                    for k, v in operations.items()
                    if isinstance(v, dict) or v._app is app
                }
            )

        return paths

    def _build_security(self):
        return [
            {sec.fields["name"]: sec.fields["value"]}
            if sec.fields["name"] is not None
            else {}
            for sec in self.security
        ]



from sanic.exceptions import SanicException

from sanic_ext.extensions.health.endpoint import setup_health_endpoint

from ..base import Extension
from .monitor import HealthMonitor


class HealthExtension(Extension):
    name = "health"
    MIN_VERSION = (22, 9)

    def startup(self, bootstrap) -> None:
        if self.config.HEALTH:
            if self.config.HEALTH_MONITOR:
                if self.MIN_VERSION > bootstrap.sanic_version:
                    min_version = ".".join(map(str, self.MIN_VERSION))
                    sanic_version = ".".join(map(str, bootstrap.sanic_version))
                    raise SanicException(
                        f"Health monitoring only works with Sanic "
                        f"v{min_version} and above. It looks like you are "
                        f"running {sanic_version}."
                    )
                HealthMonitor.setup(self.app)

            if self.config.HEALTH_ENDPOINT:
                setup_health_endpoint(self.app)

    def included(self):
        return self.config.HEALTH



import inspect

from functools import lru_cache, partial
from os.path import abspath, dirname, realpath

from sanic import Request
from sanic.blueprints import Blueprint
from sanic.config import Config
from sanic.log import logger
from sanic.response import file, html, json

from sanic_ext.config import PRIORITY
from sanic_ext.extensions.openapi.builders import (
    OperationStore,
    SpecificationBuilder,
)

from ...utils.route import (
    clean_route_name,
    get_all_routes,
    get_blueprinted_routes,
)


@lru_cache
def get_oauth2_redirect_html(version: str):
    import urllib

    response = urllib.request.urlopen(
        f"https://cdn.jsdelivr.net/npm/swagger-ui-dist@{version}/"
        "oauth2-redirect.html"
    ).read()

    return response.decode("utf-8")


def oauth2_handler(request: Request, version: str):
    return html(get_oauth2_redirect_html(version))


def blueprint_factory(config: Config):
    bp = Blueprint("openapi", url_prefix=config.OAS_URL_PREFIX)

    dir_path = dirname(realpath(__file__))
    dir_path = abspath(dir_path + "/ui")

    for ui in ("redoc", "swagger"):
        if getattr(config, f"OAS_UI_{ui}".upper()):
            path = getattr(config, f"OAS_PATH_TO_{ui}_HTML".upper())
            uri = getattr(config, f"OAS_URI_TO_{ui}".upper())
            version = getattr(config, f"OAS_UI_{ui}_VERSION".upper(), "")
            html_title = getattr(config, f"OAS_UI_{ui}_HTML_TITLE".upper())
            custom_css = getattr(config, f"OAS_UI_{ui}_CUSTOM_CSS".upper())
            html_path = path if path else f"{dir_path}/{ui}.html"

            with open(html_path, "r") as f:
                page = f.read()

            def index(
                request: Request, page: str, html_title: str, custom_css: str
            ):
                prefix = (
                    request.app.url_for("openapi.index", _external=True)
                    if getattr(request.app.config, "SERVER_NAME", None)
                    else getattr(request.app.config, "OAS_URL_PREFIX")
                ).rstrip("/")
                return html(
                    page.replace("__VERSION__", version)
                    .replace("__URL_PREFIX__", prefix)
                    .replace("__HTML_TITLE__", html_title)
                    .replace("__HTML_CUSTOM_CSS__", custom_css)
                )

            bp.add_route(
                partial(
                    index,
                    page=page,
                    html_title=html_title,
                    custom_css=custom_css,
                ),
                uri,
                name=ui,
            )
            if config.OAS_UI_DEFAULT and config.OAS_UI_DEFAULT == ui:
                bp.add_route(
                    partial(
                        index,
                        page=page,
                        html_title=html_title,
                        custom_css=custom_css,
                    ),
                    "",
                    name="index",
                )

            if ui == "swagger":
                oauth2_redirect_uri = getattr(
                    config, "OAS_UI_SWAGGER_OAUTH2_REDIRECT"
                )

                bp.add_route(
                    partial(oauth2_handler, version=version),
                    oauth2_redirect_uri,
                    name="oauth2-redirect",
                )

    @bp.get(config.OAS_URI_TO_JSON)
    async def spec(request: Request):
        if config.OAS_CUSTOM_FILE:
            return await file(config.OAS_CUSTOM_FILE)
        return json(SpecificationBuilder().build(request.app).serialize())

    if config.OAS_UI_SWAGGER:

        @bp.get(config.OAS_URI_TO_CONFIG)
        def openapi_config(request: Request):
            return json(request.app.config.SWAGGER_UI_CONFIGURATION)

    @bp.before_server_start(priority=PRIORITY)
    def build_spec(app, loop):
        specification = SpecificationBuilder()
        # --------------------------------------------------------------- #
        # Blueprint Tags
        # --------------------------------------------------------------- #

        for blueprint_name, handler in get_blueprinted_routes(app):
            operation = OperationStore()[handler]
            if not operation.tags:
                operation.tag(blueprint_name)

        # --------------------------------------------------------------- #
        # Operations
        # --------------------------------------------------------------- #
        for (
            uri,
            route_name,
            route_parameters,
            method_handlers,
            host,
        ) in get_all_routes(app, bp.url_prefix):
            # --------------------------------------------------------------- #
            # Methods
            # --------------------------------------------------------------- #

            for method, _handler in method_handlers:
                if (
                    (method == "OPTIONS" and app.config.OAS_IGNORE_OPTIONS)
                    or (method == "HEAD" and app.config.OAS_IGNORE_HEAD)
                    or method == "TRACE"
                ):
                    continue

                if hasattr(_handler, "view_class"):
                    _handler = getattr(_handler.view_class, method.lower())
                store = OperationStore()
                if (
                    _handler not in store
                    and (func := getattr(_handler, "__func__", None))
                    and func in store
                ):
                    _handler = func
                operation = store[_handler]

                if operation._exclude or "openapi" in operation.tags:
                    continue

                docstring = inspect.getdoc(_handler)

                if (
                    docstring
                    and app.config.OAS_AUTODOC
                    and operation._allow_autodoc
                ):
                    operation.autodoc(docstring)

                operation._default[
                    "operationId"
                ] = f"{method.lower()}~{route_name}"
                operation._default["summary"] = clean_route_name(route_name)

                if host:
                    if "servers" not in operation._default:
                        operation._default["servers"] = []
                    operation._default["servers"].append({"url": f"//{host}"})

                for _parameter in route_parameters:
                    if any(
                        (
                            param.fields["name"] == _parameter.name
                            for param in operation.parameters
                        )
                    ):
                        continue

                    kwargs = {}
                    if operation._autodoc and (
                        parameters := operation._autodoc.get("parameters")
                    ):
                        for param in parameters:
                            if param.pop("name", None) == _parameter.name:
                                kwargs["description"] = param.get(
                                    "description"
                                )
                                kwargs["required"] = param.get("required")
                                if schema := param.get("schema"):
                                    logger.warning(
                                        f"Ignoring the schema {schema} in "
                                        f"'{route_name}' for "
                                        f"'{_parameter.name}'. "
                                        "Instead of using the definition in "
                                        "docstring definition, Sanic will use "
                                        "the actual schema defined for this "
                                        "parameter on the route."
                                    )
                                break

                    operation.parameter(
                        _parameter.name, _parameter.cast, "path", **kwargs
                    )

                operation._app = app
                specification.operation(uri, method, operation)

        add_static_info_to_spec_from_config(app, specification)

    return bp


def add_static_info_to_spec_from_config(app, specification):
    """
    Reads app.config and sets attributes to specification according to the
    desired values.

    Modifies specification in-place and returns None
    """
    specification._do_describe(
        getattr(app.config, "API_TITLE", "API"),
        getattr(app.config, "API_VERSION", "1.0.0"),
        getattr(app.config, "API_DESCRIPTION", None),
        getattr(app.config, "API_TERMS_OF_SERVICE", None),
    )

    specification._do_license(
        getattr(app.config, "API_LICENSE_NAME", None),
        getattr(app.config, "API_LICENSE_URL", None),
    )

    specification._do_contact(
        getattr(app.config, "API_CONTACT_NAME", None),
        getattr(app.config, "API_CONTACT_URL", None),
        getattr(app.config, "API_CONTACT_EMAIL", None),
    )

    schemes = getattr(app.config, "API_SCHEMES", ["http"]) or ["http"]
    if isinstance(schemes, str):
        schemes = [s.strip() for s in schemes.split(",")]
    for scheme in schemes:
        host = getattr(app.config, "API_HOST", None)
        basePath = getattr(app.config, "API_BASEPATH", "")
        if host is None or basePath is None:
            continue

        specification.url(f"{scheme}://{host}/{basePath}")



from sanic import Request, Sanic, text

from sanic_ext.extensions.openapi import openapi
from sanic_ext.extensions.openapi.definitions import Parameter

from .utils import get_spec


def test_parameter(app: Sanic):
    DESCRIPTION = "val1 path param"
    NAME = "val1"
    LOCATION = "path"
    TYPE = "integer"

    @app.route("/test1/<val1>")
    async def handler1(request: Request, val1: int):
        """
        openapi:
        ---
        operationId: get.test1
        parameters:
          - name: val1
            in: path
            description: val1 path param
            required: true
            schema:
                type: integer
                format: int32
        """
        return text("ok")

    @app.route("/test2/<val1>")
    @openapi.parameter(
        parameter=Parameter(
            name="val1",
            schema=int,
            location=LOCATION,
            description=DESCRIPTION,
            required=True,
        )
    )
    async def handler2(request: Request, val1: int):
        return text("ok")

    @app.route("/test3/<val1>")
    @openapi.parameter(
        "val1",
        description=DESCRIPTION,
        required=True,
        schema=int,
        location=LOCATION,
    )
    async def handler3(request: Request, val1: int):
        return text("ok")

    @app.route("/test4/<val1>")
    @openapi.definition(
        parameter={
            "name": "val1",
            "description": DESCRIPTION,
            "required": True,
            "schema": int,
            "location": LOCATION,
        }
    )
    async def handler4(request: Request, val1: int):
        return text("ok")

    @app.route("/test5/<val1>")
    @openapi.definition(
        parameter=Parameter(
            name="val1",
            schema=int,
            location=LOCATION,
            description=DESCRIPTION,
            required=True,
        )
    )
    async def handler5(request: Request, val1: int):
        return text("ok")

    @app.route("/test6/<val1:strorempty>")
    async def handler6(request: Request, val1: str):
        """
        openapi:
        ---
        operationId: get.test1
        parameters:
          - name: val1
            in: path
            description: val1 path param
            required: false
        """
        return text("ok")

    @app.route("/test7/<val1:ext>")
    @openapi.parameter(
        parameter=Parameter(
            name="val1",
            location=LOCATION,
            description=DESCRIPTION,
            required=True,
        )
    )
    async def handler7(request: Request, val1: int):
        return text("ok")

    @app.route("/test8/<val1:alpha>")
    @openapi.parameter(
        parameter=Parameter(
            name="val1",
            location=LOCATION,
            description=DESCRIPTION,
            required=True,
        )
    )
    async def handler8(request: Request, val1: int):
        return text("ok")

    @app.route("/test9/<val1:slug>")
    @openapi.parameter(
        parameter=Parameter(
            name="val1",
            location=LOCATION,
            description=DESCRIPTION,
            required=True,
        )
    )
    async def handler9(request: Request, val1: int):
        return text("ok")

    spec = get_spec(app)
    for i in range(1, 6):
        assert f"/test{i}/{{val1}}" in spec["paths"]
        parameter = spec["paths"][f"/test{i}/{{val1}}"]["get"]["parameters"][0]
        assert parameter["name"] == NAME
        assert parameter["in"] == LOCATION
        assert parameter["required"] is True
        assert parameter["schema"]["type"] == TYPE
        assert parameter["description"] == DESCRIPTION

    for i in range(6, 10):
        assert f"/test{i}/{{val1}}" in spec["paths"]
        parameter = spec["paths"][f"/test{i}/{{val1}}"]["get"]["parameters"][0]
        assert parameter["name"] == NAME
        assert parameter["in"] == LOCATION
        assert parameter["required"] is not (i == 6)
        assert parameter["schema"]["type"] == "string"
        assert parameter["description"] == DESCRIPTION



from __future__ import annotations

from unittest.mock import AsyncMock, Mock, call

import pytest
from sanic import Request
from sanic.exceptions import ServerError

from sanic_ext.exceptions import InitError
from sanic_ext.extensions.injection.constructor import Constructor, gather_args
from sanic_ext.extensions.injection.registry import (
    ConstantRegistry,
    InjectionRegistry,
)


class Foo:
    def __init__(self, num: int):
        self.num = num
        self.mock = AsyncMock()

    @classmethod
    async def create(cls, request: Request, bar: int):
        instance = cls(bar)
        await instance.mock(request, bar)
        return instance


class Bar:
    def __init__(self, foo: Foo):
        self.foo = foo

    @classmethod
    async def create(cls, request: Request, foo: Foo):
        return cls(foo)


class Chicken:
    @classmethod
    def create(cls, egg: Egg):
        return cls()


class Egg:
    @classmethod
    def create(cls, chicken: Chicken):
        return cls()


class InheritedRequest(Request):
    ...


class Baz:
    @classmethod
    async def create(cls, request: InheritedRequest):
        return cls()


@pytest.mark.asyncio
async def test_gather_args():
    func = AsyncMock()
    func.return_value = 999
    injections = {"some_foo": (Foo, func)}
    request = object()
    kwargs = {"zero": 0, "one": True, "two": "2", "three": None}

    args = await gather_args(injections, request, **kwargs)
    assert args == {"some_foo": 999}
    assert func.call_args == call(request, **kwargs)


@pytest.mark.asyncio
async def test_circular_refs():
    injections = InjectionRegistry()
    injections.register(Chicken, Chicken.create)
    injections.register(Egg, Egg.create)

    assert isinstance(injections[Chicken], Constructor)
    assert isinstance(injections[Egg], Constructor)

    message = (
        "Circular dependency injection detected on 'create'. Check "
        "dependencies of 'create' which may contain circular dependency "
        f"chain with {Chicken}."
    )
    with pytest.raises(InitError, match=message):
        injections.finalize(Mock(), ConstantRegistry({}), [])


@pytest.mark.asyncio
@pytest.mark.parametrize("raises,allowed", ((False, (int,)), (True, [])))
async def test_finalize_allowed_types(raises, allowed):
    injections = InjectionRegistry()
    injections.register(Foo, Foo.create)

    if raises:
        with pytest.raises(
            InitError, match=".Could not find the following dependencies."
        ):
            injections.finalize(Mock(), ConstantRegistry({}), allowed)
    else:
        injections.finalize(Mock(), ConstantRegistry({}), allowed)


@pytest.mark.asyncio
async def test_constructor():
    injections = InjectionRegistry()
    injections.register(Foo, Foo.create)
    injections.finalize(Mock(), ConstantRegistry({}), {int})
    request = object()
    constructor = injections[Foo]

    foo = await constructor(request, bar=999)

    assert isinstance(foo, Foo)
    assert foo.num == 999
    assert constructor.pass_kwargs
    foo.mock.assert_awaited_once_with(request, 999)


@pytest.mark.asyncio
async def test_constructor_nested():
    injections = InjectionRegistry()
    injections.register(Foo, Foo.create)
    injections.register(Bar, Bar.create)
    injections.finalize(Mock(), ConstantRegistry({}), {int})
    request = object()
    constructor_bar = injections[Bar]

    bar = await constructor_bar(request, bar=999)

    assert isinstance(bar, Bar)
    assert isinstance(bar.foo, Foo)
    assert bar.foo.num == 999
    assert not constructor_bar.pass_kwargs
    bar.foo.mock.assert_awaited_once_with(request, 999)


@pytest.mark.asyncio
async def test_constructor_failure_kwargs():
    injections = InjectionRegistry()
    injections.register(Foo, Foo.create)
    injections.finalize(Mock(), ConstantRegistry({}), {int})
    request = object()
    constructor = injections[Foo]

    message = (
        "Failure to inject dependencies. Make sure that all dependencies "
        "for 'create' have been registered."
    )
    with pytest.raises(ServerError, match=message):
        await constructor(request)


@pytest.mark.asyncio
async def test_constructor_failure_nested():
    injections = InjectionRegistry()
    injections.register(Foo, Foo.create)
    injections.register(Bar, Bar.create)
    injections.finalize(Mock(), ConstantRegistry({}), {int})
    request = object()
    constructor: Bar = injections[Bar]

    message = (
        "Failure to inject dependencies. Make sure that all dependencies "
        "for 'create' have been registered."
    )
    with pytest.raises(ServerError, match=message):
        await constructor(request)


@pytest.mark.asyncio
async def test_constructor_with_inherited_request():
    injections = InjectionRegistry()
    injections.register(Baz, Baz.create)
    injections.finalize(Mock(), ConstantRegistry({}), [])

    request = object()
    constructor_baz = injections[Baz]

    baz = await constructor_baz(request)

    assert isinstance(baz, Baz)
    assert not constructor_baz.pass_kwargs



import sys
from dataclasses import dataclass, field
from typing import Dict, List, Literal, Optional, Union


@dataclass
class ModelStr:
    foo: str


@dataclass
class ModelInt:
    foo: int


@dataclass
class ModelFloat:
    foo: float


@dataclass
class ModelBool:
    foo: bool


@dataclass
class ModelOptionalStr:
    foo: Optional[str]


@dataclass
class ModelUnion:
    foo: Union[int, float]


@dataclass
class ModelUnionModels:
    foo: Union[ModelInt, ModelFloat]


@dataclass
class ModelUnionStrInt:
    foo: Union[str, int]


@dataclass
class ModelUnionIntStr:
    foo: Union[int, str]


@dataclass
class ModelOptionalUnionStrInt:
    foo: Optional[Union[str, int]]


@dataclass
class ModelOptionalUnionIntStr:
    foo: Optional[Union[int, str]]


@dataclass
class ModelListStr:
    foo: List[str]


@dataclass
class ModelListModel:
    foo: List[ModelStr]


@dataclass
class ModelOptionalList:
    foo: Optional[List[str]]


@dataclass
class ModelListUnion:
    foo: List[Union[int, float]]


@dataclass
class ModelOptionalListUnion:
    foo: Optional[List[Union[int, float]]]


@dataclass
class ModelModel:
    foo: ModelStr


@dataclass
class ModelOptionalModel:
    foo: Optional[ModelStr]


@dataclass
class ModelDictStr:
    foo: Dict[str, str]


@dataclass
class ModelDictModel:
    foo: Dict[str, ModelStr]


@dataclass
class ModelOptionalDict:
    foo: Optional[Dict[str, str]]


@dataclass
class ModelDictUnion:
    foo: Dict[str, Union[int, float]]


@dataclass
class ModelOptionalDictUnion:
    foo: Optional[Dict[str, Union[int, float]]]


@dataclass
class ModelSingleLiteral:
    foo: Literal[True]


@dataclass
class ModelMultipleLiteral:
    foo: Literal[True, "y", "Y", 1]


@dataclass
class ModelOptionalSingleLiteral:
    foo: Optional[Literal[True]]


@dataclass
class ModelOptionalMultipleLiteral:
    foo: Optional[Literal[True, "y", "Y", 1]]


@dataclass
class ModelListStrWithDefaultFactory:
    foo: List[str] = field(default_factory=list)


if sys.version_info > (3, 10):

    @dataclass
    class ModelUnionTypeStrNone:
        foo: str | None

    @dataclass
    class ModelUnionTypeStrIntNone:
        foo: str | int | None

    @dataclass
    class ModelUnionTypeStrInt:
        foo: str | int



from functools import wraps
from inspect import isawaitable, signature
from typing import Callable, TypeVar

from sanic import response


T = TypeVar("T")


def serializer(func, *, status: int = 200) -> Callable[[T], T]:
    sig = signature(func)
    simple = len(sig.parameters) == 2 or (
        func
        in (
            response.HTTPResponse,
            response.file_stream,
            response.file,
            response.html,
            response.json,
            response.raw,
            response.redirect,
            response.text,
        )
    )

    def decorator(f):
        @wraps(f)
        async def decorated_function(*args, **kwargs):
            retval = f(*args, **kwargs)
            if isawaitable(retval):
                retval = await retval

            if simple:
                return func(retval, status=status)
            else:
                kwargs["status"] = status
                return func(retval, *args, **kwargs)

        return decorated_function

    return decorator



import pytest
from sanic import Sanic
from sanic.response import empty

from sanic_ext.extensions.logging.logger import Logger


def test_defult_config(app: Sanic):
    assert hasattr(app.config, "LOGGERS")


def test_custom_background_logger(app: Sanic):
    assert Logger.LOGGERS == []
    Logger.prepare(app)
    assert hasattr(app.shared_ctx, "logger_queue")
    assert Logger.LOGGERS == app.config.LOGGERS
    assert "sanic.root" in Logger.LOGGERS

    logger = Logger()
    assert (
        len(logger.loggers) == len(Logger.LOGGERS) == len(app.config.LOGGERS)
    )



from __future__ import annotations

from dataclasses import _HAS_DEFAULT_FACTORY  # type: ignore
from typing import (
    Any,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    Union,
    get_args,
    get_origin,
)

from sanic_ext.utils.typing import (
    UnionType,
    is_generic,
    is_msgspec,
    is_optional,
)


MISSING: Tuple[Any, ...] = (_HAS_DEFAULT_FACTORY,)

try:
    import attrs  # noqa

    NOTHING = attrs.NOTHING
    ATTRS = True
    MISSING = (
        _HAS_DEFAULT_FACTORY,
        NOTHING,
    )
except ImportError:
    ATTRS = False


try:
    import msgspec

    MSGSPEC = True
except ImportError:
    MSGSPEC = False


class Hint(NamedTuple):
    hint: Any
    model: bool
    literal: bool
    typed: bool
    nullable: bool
    origin: Optional[Any]
    allowed: Tuple[Hint, ...]  # type: ignore
    allow_missing: bool

    def validate(
        self, value, schema, allow_multiple=False, allow_coerce=False
    ):
        if not self.typed:
            if self.model:
                return check_data(
                    self.hint,
                    value,
                    schema,
                    allow_multiple=allow_multiple,
                    allow_coerce=allow_coerce,
                )

            if (
                allow_multiple
                and isinstance(value, list)
                and self.coerce_type is not list
                and len(value) == 1
            ):
                value = value[0]
            try:
                _check_types(value, self.literal, self.hint)
            except ValueError as e:
                if allow_coerce:
                    value = self.coerce(value)
                    _check_types(value, self.literal, self.hint)
                else:
                    raise e
        else:
            value = _check_nullability(
                value,
                self.nullable,
                self.allowed,
                schema,
                allow_multiple,
                allow_coerce,
            )

            if not self.nullable:
                if self.origin in (Union, Literal, UnionType):
                    value = _check_inclusion(
                        value,
                        self.allowed,
                        schema,
                        allow_multiple,
                        allow_coerce,
                    )
                elif self.origin is list:
                    value = _check_list(
                        value,
                        self.allowed,
                        self.hint,
                        schema,
                        allow_multiple,
                        allow_coerce,
                    )
                elif self.origin is dict:
                    value = _check_dict(
                        value,
                        self.allowed,
                        self.hint,
                        schema,
                        allow_multiple,
                        allow_coerce,
                    )

            if allow_coerce:
                value = self.coerce(value)

        return value

    def coerce(self, value):
        if is_generic(self.coerce_type):
            args = get_args(self.coerce_type)
            if get_origin(self.coerce_type) == Literal or (
                all(get_origin(arg) == Literal for arg in args)
            ):
                return value
            if type(None) in args and value is None:
                return None
            coerce_types = [arg for arg in args if not isinstance(None, arg)]
        else:
            coerce_types = [self.coerce_type]
        for coerce_type in coerce_types:
            try:
                if isinstance(value, list):
                    value = [coerce_type(item) for item in value]
                else:
                    value = coerce_type(value)
            except (ValueError, TypeError):
                ...
            else:
                return value
        return value

    @property
    def coerce_type(self):
        coerce_type = self.hint
        if is_optional(coerce_type):
            coerce_type = get_args(self.hint)[0]
        return coerce_type


def check_data(model, data, schema, allow_multiple=False, allow_coerce=False):
    if not isinstance(data, dict):
        raise TypeError(f"Value '{data}' is not a dict")
    sig = schema[model.__name__]["sig"]
    hints = schema[model.__name__]["hints"]
    bound = sig.bind(**data)
    bound.apply_defaults()
    params = dict(zip(sig.parameters, bound.args))
    params.update(bound.kwargs)

    hydration_values = {}
    try:
        for key, value in params.items():
            hint = hints.get(key, Any)
            try:
                hydration_values[key] = hint.validate(
                    value,
                    schema,
                    allow_multiple=allow_multiple,
                    allow_coerce=allow_coerce,
                )
            except ValueError:
                if not hint.allow_missing or value not in MISSING:
                    raise
    except ValueError as e:
        raise TypeError(e)

    if MSGSPEC and is_msgspec(model):
        try:
            return msgspec.from_builtins(
                hydration_values, model, str_values=True, str_keys=True
            )
        except msgspec.ValidationError as e:
            raise TypeError(e)
    else:
        return model(**hydration_values)


def _check_types(value, literal, expected):
    if literal:
        if expected is Any:
            return
        elif value != expected:
            raise ValueError(f"Value '{value}' must be {expected}")
    else:
        if MSGSPEC and is_msgspec(expected) and isinstance(value, Mapping):
            try:
                expected(**value)
            except (TypeError, msgspec.ValidationError):
                raise ValueError(f"Value '{value}' is not of type {expected}")
        elif not isinstance(value, expected):
            raise ValueError(f"Value '{value}' is not of type {expected}")


def _check_nullability(
    value, nullable, allowed, schema, allow_multiple, allow_coerce
):
    if not nullable and value is None:
        raise ValueError("Value cannot be None")
    if nullable and value is not None:
        exc = None
        for hint in allowed:
            try:
                value = hint.validate(
                    value, schema, allow_multiple, allow_coerce
                )
            except ValueError as e:
                exc = e
            else:
                break
        else:
            if exc:
                if len(allowed) == 1:
                    raise exc
                else:
                    options = ", ".join(
                        [str(option.hint) for option in allowed]
                    )
                    raise ValueError(
                        f"Value '{value}' must be one of {options}, or None"
                    )
    return value


def _check_inclusion(value, allowed, schema, allow_multiple, allow_coerce):
    for option in allowed:
        try:
            return option.validate(value, schema, allow_multiple, allow_coerce)
        except (ValueError, TypeError):
            ...

    options = ", ".join([str(option.hint) for option in allowed])
    raise ValueError(f"Value '{value}' must be one of {options}")


def _check_list(value, allowed, hint, schema, allow_multiple, allow_coerce):
    if isinstance(value, list):
        try:
            return [
                _check_inclusion(
                    item, allowed, schema, allow_multiple, allow_coerce
                )
                for item in value
            ]
        except (ValueError, TypeError):
            ...
    raise ValueError(f"Value '{value}' must be a {hint}")


def _check_dict(value, allowed, hint, schema, allow_multiple, allow_coerce):
    if isinstance(value, dict):
        try:
            return {
                key: _check_inclusion(
                    item, allowed, schema, allow_multiple, allow_coerce
                )
                for key, item in value.items()
            }
        except (ValueError, TypeError):
            ...
    raise ValueError(f"Value '{value}' must be a {hint}")



from datetime import date, datetime
from enum import Enum
from typing import Any, Dict, Optional, Union
from uuid import UUID

import pytest
from sanic.exceptions import SanicException
from sanic.views import HTTPMethodView

from sanic_ext import openapi

from .utils import get_path, get_spec


class Choice(Enum):
    ONE = 1
    TWO = 2


class Bar:
    name: str


class LittleFoo:
    iid: int
    bar: Bar


class BigFoo:
    iid: int
    uid: UUID
    created: date
    updated: datetime
    multi: Union[str, int]
    nullable_single: Optional[bool]
    nullable_multi: Optional[Union[str, int]]
    bar: Bar
    adict: Dict[str, Any]
    bdict: Dict[str, bool]
    anything: Any
    choice: Choice


@pytest.mark.parametrize(
    "args,content_type",
    (
        ((LittleFoo,), "*/*"),
        (({"application/json": LittleFoo},), "application/json"),
        ((openapi.definitions.RequestBody(LittleFoo),), "*/*"),
        (
            (
                openapi.definitions.RequestBody(
                    {"application/json": LittleFoo}
                ),
            ),
            "application/json",
        ),
    ),
)
def test_body_decorator(app, args, content_type):
    @app.route("/")
    @openapi.body(*args, description="something")
    async def handler(_):
        ...

    spec = get_path(app, "/")

    assert spec["requestBody"]["description"] == "something"

    content = spec["requestBody"]["content"]

    assert content_type in content

    schema = content[content_type]["schema"]

    assert schema["type"] == "object"
    assert schema["properties"]["bar"]["type"] == "object"
    assert (
        schema["properties"]["bar"]["properties"]["name"]["type"] == "string"
    )


@pytest.mark.parametrize(
    "decorator", (openapi.deprecated(), openapi.deprecated)
)
def test_deprecated_decorator(app, decorator):
    @app.route("/")
    @decorator
    async def handler(_):
        ...

    spec = get_path(app, "/")
    assert spec["deprecated"]


def test_description_decorator(app):
    @app.route("/")
    @openapi.description("foo")
    async def handler(_):
        ...

    spec = get_path(app, "/")
    assert spec["description"] == "foo"


@pytest.mark.parametrize(
    "args",
    (
        ("http://example.com/docs",),
        (
            openapi.definitions.ExternalDocumentation(
                "http://example.com/docs"
            ),
        ),
    ),
)
def test_document_decorator(app, args):
    @app.route("/")
    @openapi.document(*args)
    async def handler(_):
        ...

    spec = get_path(app, "/")
    assert spec["externalDocs"]["url"] == "http://example.com/docs"


@pytest.mark.parametrize(
    "decorator,excluded",
    (
        (openapi.exclude(), True),
        (openapi.exclude(True), True),
        (openapi.exclude(False), False),
    ),
)
def test_exclude_decorator(app, decorator, excluded):
    @app.route("/")
    @decorator
    async def handler(_):
        ...

    if excluded:
        with pytest.raises(KeyError):
            get_path(app, "/")
    else:
        get_path(app, "/")


def test_operation_decorator(app):
    @app.route("/")
    @openapi.operation("foo")
    async def handler(_):
        ...

    spec = get_path(app, "/")
    assert spec["operationId"] == "foo"


@pytest.mark.parametrize(
    "decorator,expected",
    (
        (
            openapi.parameter("thing"),
            {
                "name": "thing",
                "schema": {"type": "string"},
                "in": "query",
            },
        ),
        (
            openapi.parameter("Authorization", str, "header"),
            {
                "name": "Authorization",
                "schema": {"type": "string"},
                "in": "header",
            },
        ),
        (
            openapi.parameter(
                parameter=openapi.definitions.Parameter(
                    "foobar", deprecated=True
                )
            ),
            {
                "name": "foobar",
                "schema": {"type": "string"},
                "deprecated": True,
                "in": "query",
            },
        ),
        (
            openapi.parameter("thing", required=True),
            {
                "name": "thing",
                "schema": {"type": "string"},
                "required": True,
                "in": "query",
            },
        ),
    ),
)
def test_parameter_decorator(app, decorator, expected):
    @app.route("/")
    @decorator
    def handler_one(_):
        ...

    parameters = get_path(app, "/")["parameters"]
    assert parameters[0] == expected


@pytest.mark.parametrize(
    "decorator,expected",
    (
        (
            openapi.response(),
            {
                "default": {
                    "content": {"*/*": {"schema": {"type": "string"}}},
                    "description": "Default Response",
                }
            },
        ),
        (
            openapi.response(200, str, "foobar"),
            {
                "200": {
                    "content": {"*/*": {"schema": {"type": "string"}}},
                    "description": "foobar",
                }
            },
        ),
        (
            openapi.response(200, Bar, "..."),
            {
                "200": {
                    "content": {
                        "*/*": {
                            "schema": {
                                "type": "object",
                                "properties": {"name": {"type": "string"}},
                            }
                        }
                    },
                    "description": "...",
                }
            },
        ),
        (
            openapi.response(
                content={"application/json": Bar}, description="..."
            ),
            {
                "default": {
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {"name": {"type": "string"}},
                            }
                        }
                    },
                    "description": "...",
                }
            },
        ),
        (
            openapi.response(
                response=openapi.definitions.Response(
                    {"application/json": Bar}, description="...", status=201
                )
            ),
            {
                "201": {
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {"name": {"type": "string"}},
                            }
                        }
                    },
                    "description": "...",
                }
            },
        ),
        (
            openapi.response(content={"application/json": BigFoo}),
            {
                "default": {
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {
                                    "iid": {
                                        "type": "integer",
                                        "format": "int32",
                                    },
                                    "uid": {
                                        "type": "string",
                                        "format": "uuid",
                                    },
                                    "created": {
                                        "type": "string",
                                        "format": "date",
                                    },
                                    "updated": {
                                        "type": "string",
                                        "format": "date-time",
                                    },
                                    "multi": {
                                        "oneOf": [
                                            {"type": "string"},
                                            {
                                                "type": "integer",
                                                "format": "int32",
                                            },
                                        ]
                                    },
                                    "nullable_single": {
                                        "type": "boolean",
                                        "nullable": True,
                                    },
                                    "nullable_multi": {
                                        "nullable": True,
                                        "oneOf": [
                                            {
                                                "type": "string",
                                            },
                                            {
                                                "type": "integer",
                                                "format": "int32",
                                            },
                                        ],
                                    },
                                    "bar": {
                                        "type": "object",
                                        "properties": {
                                            "name": {"type": "string"}
                                        },
                                    },
                                    "adict": {
                                        "type": "object",
                                        "additionalProperties": {},
                                    },
                                    "bdict": {
                                        "type": "object",
                                        "additionalProperties": {
                                            "type": "boolean"
                                        },
                                    },
                                    "anything": {},
                                    "choice": {
                                        "type": "integer",
                                        "format": "int32",
                                        "enum": [1, 2],
                                    },
                                },
                            }
                        }
                    },
                    "description": "Default Response",
                }
            },
        ),
    ),
)
def test_response_decorator(app, decorator, expected):
    @app.route("/")
    @decorator
    def handler_one(_):
        ...

    responses = get_path(app, "/")["responses"]
    assert responses == expected


def test_summary_decorator(app):
    @app.route("/")
    @openapi.summary("foo")
    async def handler(_):
        ...

    spec = get_path(app, "/")
    assert spec["summary"] == "foo"


@pytest.mark.parametrize(
    "decorator,tags",
    (
        (openapi.tag("foo"), ("foo",)),
        (openapi.tag("foo", openapi.definitions.Tag("bar")), ("foo", "bar")),
    ),
)
def test_tag_decorator(app, decorator, tags):
    @app.route("/")
    @decorator
    def handler_one(_):
        ...

    spec = get_spec(app)
    for tag in tags:
        assert {"name": tag} in spec["tags"]

    tagged = get_path(app, "/")["tags"]
    assert list(tagged) == list(tags)


def test_definition_decorator_body_dict_w_obj(app):
    @app.route("/")
    @openapi.definition(body={"application/json": Bar})
    async def handler(_):
        ...

    body = get_path(app, "/")["requestBody"]
    assert body == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        }
    }


def test_definition_decorator_body_dict_only_schema_head(app):
    @app.route("/")
    @openapi.definition(
        body={
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        }
    )
    async def handler(_):
        ...

    body = get_path(app, "/")["requestBody"]
    assert body == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        }
    }


def test_definition_decorator_body_dict_types_schema_head(app):
    @app.route("/")
    @openapi.definition(body={"application/json": {"schema": {"name": str}}})
    async def handler(_):
        ...

    body = get_path(app, "/")["requestBody"]
    assert body == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        }
    }


def test_definition_decorator_body_dict_only_schema_root(app):
    @app.route("/")
    @openapi.definition(
        body={
            "application/json": {
                "type": "object",
                "properties": {"name": {"type": "string"}},
            }
        }
    )
    async def handler(_):
        ...

    body = get_path(app, "/")["requestBody"]
    assert body == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        }
    }


def test_definition_decorator_body_dict_types_schema_root(app):
    @app.route("/")
    @openapi.definition(body={"application/json": {"name": str}})
    async def handler(_):
        ...

    body = get_path(app, "/")["requestBody"]
    assert body == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        }
    }


def test_definition_decorator_httpmethodview(app):
    class View(HTTPMethodView, uri="/", attach=app):
        @openapi.definition(body={"application/json": Bar})
        async def get(self, request):
            ...

    body = get_path(app, "/")["requestBody"]
    assert body == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        }
    }


def test_definition_decorator_body_dict_multi(app):
    @app.route("/")
    @openapi.definition(body={"application/json": Bar, "text/plain": str})
    async def handler(_):
        ...

    body = get_path(app, "/")["requestBody"]
    assert body == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            },
            "text/plain": {"schema": {"type": "string"}},
        }
    }


def test_definition_decorator_body_request_body(app):
    @app.route("/")
    @openapi.definition(
        body=openapi.definitions.RequestBody(str, required=True)
    )
    async def handler(_):
        ...

    body = get_path(app, "/")["requestBody"]
    assert body == {
        "content": {"*/*": {"schema": {"type": "string"}}},
        "required": True,
    }


def test_definition_decorator_body_model(app):
    @app.route("/")
    @openapi.definition(body=Bar)
    async def handler(_):
        ...

    body = get_path(app, "/")["requestBody"]
    assert body == {
        "content": {
            "*/*": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        }
    }


def test_definition_decorator_deprecated(app):
    @app.route("/")
    @openapi.definition(deprecated=True)
    async def handler(_):
        ...

    assert get_path(app, "/")["deprecated"]


def test_definition_decorator_description(app):
    @app.route("/")
    @openapi.definition(description="foo")
    async def handler(_):
        ...

    assert get_path(app, "/")["description"] == "foo"


def test_definition_decorator_document_string(app):
    @app.route("/")
    @openapi.definition(document="foo")
    async def handler(_):
        ...

    assert get_path(app, "/")["externalDocs"] == {"url": "foo"}


def test_definition_decorator_document_obj(app):
    @app.route("/")
    @openapi.definition(
        document=openapi.definitions.ExternalDocumentation("foo", "bar")
    )
    async def handler(_):
        ...

    assert get_path(app, "/")["externalDocs"] == {
        "url": "foo",
        "description": "bar",
    }


def test_definition_decorator_operation(app):
    @app.route("/")
    @openapi.definition(operation="foo")
    async def handler(_):
        ...

    assert get_path(app, "/")["operationId"] == "foo"


def test_definition_decorator_parameter_string(app):
    @app.route("/")
    @openapi.definition(parameter="foo")
    async def handler(_):
        ...

    parameters = get_path(app, "/")["parameters"]
    assert {
        "name": "foo",
        "schema": {"type": "string"},
        "in": "query",
    } in parameters


def test_definition_decorator_parameter_dict(app):
    @app.route("/")
    @openapi.definition(parameter={"name": "foo"})
    async def handler(_):
        ...

    parameters = get_path(app, "/")["parameters"]
    assert {
        "name": "foo",
        "schema": {"type": "string"},
        "in": "query",
    } in parameters


def test_definition_decorator_parameter_obj(app):
    @app.route("/")
    @openapi.definition(
        parameter=openapi.definitions.Parameter("foo", description="bar")
    )
    async def handler(_):
        ...

    parameters = get_path(app, "/")["parameters"]
    assert {
        "name": "foo",
        "schema": {"type": "string"},
        "in": "query",
        "description": "bar",
    } in parameters


def test_definition_decorator_parameter_string_multi(app):
    @app.route("/")
    @openapi.definition(parameter=["foo", "bar"])
    async def handler(_):
        ...

    parameters = get_path(app, "/")["parameters"]
    assert {
        "name": "foo",
        "schema": {"type": "string"},
        "in": "query",
    } in parameters
    assert {
        "name": "bar",
        "schema": {"type": "string"},
        "in": "query",
    } in parameters


def test_definition_decorator_parameter_dict_multi(app):
    @app.route("/")
    @openapi.definition(
        parameter=[{"name": "foo"}, {"name": "bar", "location": "header"}]
    )
    async def handler(_):
        ...

    parameters = get_path(app, "/")["parameters"]
    assert {
        "name": "foo",
        "schema": {"type": "string"},
        "in": "query",
    } in parameters
    assert {
        "name": "bar",
        "schema": {"type": "string"},
        "in": "header",
    } in parameters


def test_definition_decorator_parameter_obj_multi(app):
    @app.route("/")
    @openapi.definition(
        parameter=[
            openapi.definitions.Parameter("foo"),
            openapi.definitions.Parameter("bar"),
        ]
    )
    async def handler(_):
        ...

    parameters = get_path(app, "/")["parameters"]
    assert {
        "name": "foo",
        "schema": {"type": "string"},
        "in": "query",
    } in parameters
    assert {
        "name": "bar",
        "schema": {"type": "string"},
        "in": "query",
    } in parameters


def test_definition_decorator_response_dict(app):
    @app.route("/")
    @openapi.definition(response={"application/json": Bar})
    async def handler(_):
        ...

    responses = get_path(app, "/")["responses"]
    assert responses["default"] == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        },
        "description": "Default Response",
    }


def test_definition_decorator_response_obj(app):
    @app.route("/")
    @openapi.definition(
        response=openapi.definitions.Response(
            {"application/json": Bar}, status=201
        )
    )
    async def handler(_):
        ...

    responses = get_path(app, "/")["responses"]
    assert responses["201"] == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        },
        "description": "Default Response",
    }


def test_definition_decorator_response_model(app):
    @app.route("/")
    @openapi.definition(response=Bar)
    async def handler(_):
        ...

    responses = get_path(app, "/")["responses"]
    assert responses["default"] == {
        "content": {
            "*/*": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        },
        "description": "Default Response",
    }


def test_definition_decorator_response_dict_multi(app):
    @app.route("/")
    @openapi.definition(
        response=[
            {"application/json": Bar, "text/plain": str},
            {"text/html": str},
        ]
    )
    async def handler(_):
        ...

    responses = get_path(app, "/")["responses"]

    assert responses["default"] == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            },
            "text/plain": {"schema": {"type": "string"}},
            "text/html": {"schema": {"type": "string"}},
        },
        "description": "Default Response",
    }


def test_definition_decorator_response_obj_multi(app):
    @app.route("/")
    @openapi.definition(
        response=[
            openapi.definitions.Response(
                {"application/json": Bar}, status=201
            ),
            openapi.definitions.Response(
                {"*/*": str}, status=400, description="Something bad"
            ),
        ]
    )
    async def handler(_):
        ...

    responses = get_path(app, "/")["responses"]
    assert responses["201"] == {
        "content": {
            "application/json": {
                "schema": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            }
        },
        "description": "Default Response",
    }
    assert responses["400"] == {
        "content": {"*/*": {"schema": {"type": "string"}}},
        "description": "Something bad",
    }


def test_definition_decorator_response_model_multi(app):
    with pytest.raises(SanicException):

        @app.route("/")
        @openapi.definition(response=[Bar, LittleFoo])
        async def handler(_):
            ...


def test_definition_decorator_summary(app):
    @app.route("/")
    @openapi.definition(summary="foo")
    async def handler(_):
        ...

    assert get_path(app, "/")["summary"] == "foo"


def test_definition_decorator_tag_string(app):
    @app.route("/")
    @openapi.definition(tag="foo")
    async def handler(_):
        ...

    assert list(get_path(app, "/")["tags"]) == ["foo"]


def test_definition_decorator_tag_obj(app):
    @app.route("/")
    @openapi.definition(tag=openapi.definitions.Tag("foo"))
    async def handler(_):
        ...

    assert list(get_path(app, "/")["tags"]) == ["foo"]


def test_definition_decorator_tag_string_multi(app):
    @app.route("/")
    @openapi.definition(tag=["foo", "bar"])
    async def handler(_):
        ...

    assert list(get_path(app, "/")["tags"]) == ["foo", "bar"]


def test_definition_decorator_tag_obj_multi(app):
    @app.route("/")
    @openapi.definition(
        tag=[openapi.definitions.Tag("foo"), openapi.definitions.Tag("bar")]
    )
    async def handler(_):
        ...

    assert list(get_path(app, "/")["tags"]) == ["foo", "bar"]



from sanic import Request
from sanic.exceptions import SanicException


def extract_request(*args) -> Request:
    request: Request
    if args and isinstance(args[0], Request):
        request = args[0]
    elif len(args) > 1:
        request = args[1]
    else:
        raise SanicException("Request could not be found")
    return request



import logging

from collections import defaultdict
from logging import LogRecord
from logging.handlers import QueueHandler
from multiprocessing import Manager
from queue import Empty, Full
from signal import SIGINT, SIGTERM
from signal import signal as signal_func
from typing import List

from sanic import Sanic
from sanic.log import server_logger


async def prepare_logger(app: Sanic, *_):
    Logger.prepare(app)


async def setup_logger(app: Sanic, *_):
    logger = Logger()
    app.manager.manage(
        "Logger",
        logger,
        {
            "queue": app.shared_ctx.logger_queue,
        },
    )


class SanicQueueHandler(QueueHandler):
    def emit(self, record: LogRecord) -> None:
        try:
            return super().enqueue(record)
        except Full:
            server_logger.warning(
                "Background logger is full. Emitting log in process."
            )
            server_logger.handle(record)


async def setup_server_logging(app: Sanic):
    qhandler = SanicQueueHandler(app.shared_ctx.logger_queue)
    app.ctx._logger_handlers = defaultdict(list)
    app.ctx._qhandler = qhandler

    for logger_name in app.config.LOGGERS:
        logger_instance = logging.getLogger(logger_name)
        for handler in logger_instance.handlers:
            logger_instance.removeHandler(handler)
        logger_instance.addHandler(qhandler)


async def remove_server_logging(app: Sanic):
    for logger, handlers in app.ctx._logger_handlers.items():
        logger.removeHandler(app.ctx._qhandler)
        for handler in handlers:
            logger.addHandler(handler)


class Logger:
    LOGGERS = []

    def __init__(self):
        self.run = True
        self.loggers = {
            logger: logging.getLogger(logger) for logger in self.LOGGERS
        }

    def __call__(self, queue) -> None:
        signal_func(SIGINT, self.stop)
        signal_func(SIGTERM, self.stop)

        while self.run:
            try:
                record: LogRecord = queue.get(timeout=0.05)
            except Empty:
                continue
            logger = self.loggers.get(record.name)
            logger.handle(record)

    def stop(self, *_):
        if self.run:
            self.run = False

    @classmethod
    def update_cls_loggers(cls, logger_names: List[str]):
        cls.LOGGERS = logger_names

    @classmethod
    def prepare(cls, app: Sanic):
        sync_manager = Manager()
        logger_queue = sync_manager.Queue(
            maxsize=app.config.LOGGING_QUEUE_MAX_SIZE
        )
        app.shared_ctx.logger_queue = logger_queue
        cls.update_cls_loggers(app.config.LOGGERS)

    @classmethod
    def setup(cls, app: Sanic):
        app.main_process_start(prepare_logger)
        app.main_process_ready(setup_logger)
        app.before_server_start(setup_server_logging)
        app.before_server_stop(remove_server_logging)



from typing import List

import attrs
from sanic import json
from sanic.views import HTTPMethodView

from sanic_ext import validate

SNOOPY_DATA = {"name": "Snoopy", "alter_ego": ["Flying Ace", "Joe Cool"]}


def test_validate_json(app):
    @attrs.define
    class Pet:
        name: str
        alter_ego: List[str]

    @app.post("/function")
    @validate(json=Pet)
    async def handler(_, body: Pet):
        return json(
            {
                "is_pet": isinstance(body, Pet),
                "pet": {"name": body.name, "alter_ego": body.alter_ego},
            }
        )

    class MethodView(HTTPMethodView, attach=app, uri="/method"):
        decorators = [validate(json=Pet)]

        async def post(self, _, body: Pet):
            return json(
                {
                    "is_pet": isinstance(body, Pet),
                    "pet": {"name": body.name, "alter_ego": body.alter_ego},
                }
            )

    _, response = app.test_client.post("/function", json=SNOOPY_DATA)
    assert response.status == 200
    assert response.json["is_pet"]
    assert response.json["pet"] == SNOOPY_DATA

    _, response = app.test_client.post("/method", json=SNOOPY_DATA)
    assert response.status == 200
    assert response.json["is_pet"]
    assert response.json["pet"] == SNOOPY_DATA


def test_validate_form(app):
    @attrs.define
    class Pet:
        name: str
        alter_ego: List[str]

    @app.post("/function")
    @validate(form=Pet)
    async def handler(_, body: Pet):
        return json(
            {
                "is_pet": isinstance(body, Pet),
                "pet": {"name": body.name, "alter_ego": body.alter_ego},
            }
        )

    class MethodView(HTTPMethodView, attach=app, uri="/method"):
        decorators = [validate(form=Pet)]

        async def post(self, _, body: Pet):
            return json(
                {
                    "is_pet": isinstance(body, Pet),
                    "pet": {"name": body.name, "alter_ego": body.alter_ego},
                }
            )

    _, response = app.test_client.post("/function", data=SNOOPY_DATA)
    assert response.status == 200
    assert response.json["is_pet"]
    assert response.json["pet"] == SNOOPY_DATA

    _, response = app.test_client.post("/method", data=SNOOPY_DATA)
    assert response.status == 200
    assert response.json["is_pet"]
    assert response.json["pet"] == SNOOPY_DATA


def test_validate_query(app):
    @attrs.define
    class Search:
        q: str

    @app.get("/function")
    @validate(query=Search)
    async def handler(_, query: Search):
        return json({"q": query.q, "is_search": isinstance(query, Search)})

    class MethodView(HTTPMethodView, attach=app, uri="/method"):
        decorators = [validate(query=Search)]

        async def get(self, _, query: Search):
            return json({"q": query.q, "is_search": isinstance(query, Search)})

    _, response = app.test_client.get("/function", params={"q": "Snoopy"})
    assert response.status == 200
    assert response.json["is_search"]
    assert response.json["q"] == "Snoopy"

    _, response = app.test_client.get("/method", params={"q": "Snoopy"})
    assert response.status == 200
    assert response.json["is_search"]
    assert response.json["q"] == "Snoopy"



from sanic import Sanic

from sanic_ext import openapi
from tests.extensions.openapi.utils import get_path, get_spec


def test_secured(app: Sanic):
    @app.route("/one")
    async def handler1(request):
        """
        openapi:
        ---
        security:
            - foo: []
        """

    @app.route("/two")
    @openapi.secured("foo")
    @openapi.secured({"bar": []})
    @openapi.secured(baz=[])
    async def handler2(request):
        ...

    @app.route("/three")
    @openapi.definition(secured="foo")
    @openapi.definition(secured={"bar": []})
    async def handler3(request):
        ...

    spec = get_path(app, "/one")
    assert {"foo": []} in spec["security"]

    spec = get_path(app, "/two")
    assert {"foo": []} in spec["security"]
    assert {"bar": []} in spec["security"]
    assert {"baz": []} in spec["security"]

    spec = get_path(app, "/three")
    assert {"foo": []} in spec["security"]
    assert {"bar": []} in spec["security"]


def test_security_scheme(app: Sanic):
    app.ext.openapi.add_security_scheme("api_key", "apiKey")
    app.ext.openapi.add_security_scheme("token1", "http")
    app.ext.openapi.add_security_scheme(
        "token2", "http", scheme="bearer", bearer_format="JWT"
    )
    app.ext.openapi.add_security_scheme("oldschool", "http", scheme="basic")
    app.ext.openapi.add_security_scheme(
        "oa2",
        "oauth2",
        flows={
            "implicit": {
                "authorizationUrl": "http://example.com/auth",
                "scopes": {
                    "one:two": "something",
                    "three:four": "something else",
                },
            }
        },
    )

    spec = get_spec(app)
    schemes = spec["components"]["securitySchemes"]

    assert schemes["api_key"] == {
        "type": "apiKey",
        "name": "authorization",
        "in": "header",
    }
    assert schemes["token1"] == {
        "type": "http",
        "scheme": "Bearer",
    }
    assert schemes["token2"] == {
        "type": "http",
        "scheme": "bearer",
        "bearerFormat": "JWT",
    }
    assert schemes["oa2"] == {
        "type": "oauth2",
        "flows": {
            "implicit": {
                "authorizationUrl": "http://example.com/auth",
                "scopes": {
                    "one:two": "something",
                    "three:four": "something else",
                },
            }
        },
    }


def test_raw(app: Sanic):
    app.ext.openapi.raw({"security": [{}, {"foo": []}]})

    spec = get_spec(app)
    assert {} in spec["security"]
    assert {"foo": []} in spec["security"]



import pytest

from .utils import get_spec


@pytest.mark.parametrize(
    "schemes,expected",
    (
        (None, ("http://example.com/",)),
        ("https", ("https://example.com/",)),
        (
            "http,https",
            (
                "http://example.com/",
                "https://example.com/",
            ),
        ),
        (
            ["http", "https"],
            (
                "http://example.com/",
                "https://example.com/",
            ),
        ),
    ),
)
def test_config_scheme(app, schemes, expected):
    app.config.API_SCHEMES = schemes
    app.config.API_HOST = "example.com"

    spec = get_spec(app)
    urls = {s["url"] for s in spec["servers"]}

    assert urls == set(expected)



from sys import version_info
from typing import Any, Dict, List

import pytest

from sanic_ext.extras.validation.schema import parse_hint


@pytest.mark.skipif(version_info < (3, 9), reason="Not needed on 3.8")
def test_parse_generic_list():
    hint_1 = parse_hint(List[int])
    hint_2 = parse_hint(list[int])

    assert hint_1.origin == hint_2.origin
    assert hint_1.allowed == hint_2.allowed


@pytest.mark.skipif(version_info < (3, 9), reason="Not needed on 3.8")
def test_parse_generic_dict():
    hint_1 = parse_hint(Dict[str, Any])
    hint_2 = parse_hint(dict[str, Any])

    assert hint_1.origin == hint_2.origin
    assert hint_1.allowed == hint_2.allowed



import re

from dataclasses import dataclass
from datetime import timedelta
from types import SimpleNamespace
from typing import Any, FrozenSet, List, Optional, Tuple, Union

from sanic import HTTPResponse, Request, Sanic
from sanic.exceptions import SanicException
from sanic.helpers import Default, _default
from sanic.log import logger

from sanic_ext.config import PRIORITY


WILDCARD_PATTERN = re.compile(r".*")
ORIGIN_HEADER = "access-control-allow-origin"
ALLOW_HEADERS_HEADER = "access-control-allow-headers"
ALLOW_METHODS_HEADER = "access-control-allow-methods"
EXPOSE_HEADER = "access-control-expose-headers"
CREDENTIALS_HEADER = "access-control-allow-credentials"
REQUEST_METHOD_HEADER = "access-control-request-method"
REQUEST_HEADERS_HEADER = "access-control-request-headers"
MAX_AGE_HEADER = "access-control-max-age"
VARY_HEADER = "vary"


@dataclass(frozen=True)
class CORSSettings:
    allow_headers: FrozenSet[str]
    allow_methods: FrozenSet[str]
    allow_origins: Tuple[re.Pattern, ...]
    always_send: bool
    automatic_options: bool
    expose_headers: FrozenSet[str]
    max_age: str
    send_wildcard: bool
    supports_credentials: bool


def add_cors(app: Sanic) -> None:
    _setup_cors_settings(app)

    @app.on_response
    async def _add_cors_headers(request, response):
        preflight = (
            request.app.ctx.cors.automatic_options
            and request.method == "OPTIONS"
        )

        if preflight and not request.headers.get(REQUEST_METHOD_HEADER):
            logger.info(
                "No Access-Control-Request-Method header found on request. "
                "CORS headers will not be applied."
            )
            return

        _add_origin_header(request, response)

        if ORIGIN_HEADER not in response.headers:
            return

        _add_expose_header(request, response)
        _add_credentials_header(request, response)
        _add_vary_header(request, response)

        if preflight:
            _add_max_age_header(request, response)
            _add_allow_header(request, response)
            _add_methods_header(request, response)

    @app.before_server_start(priority=PRIORITY)
    async def _assign_cors_settings(app, _):
        for group in app.router.groups.values():
            _cors = SimpleNamespace()
            for route in group:
                cors = getattr(route.handler, "__cors__", None)
                if cors:
                    for key, value in cors.__dict__.items():
                        setattr(_cors, key, value)

            for route in group:
                route.ctx._cors = _cors


def cors(
    *,
    origin: Union[str, Default] = _default,
    expose_headers: Union[List[str], Default] = _default,
    allow_headers: Union[List[str], Default] = _default,
    allow_methods: Union[List[str], Default] = _default,
    supports_credentials: Union[bool, Default] = _default,
    max_age: Union[str, int, timedelta, Default] = _default,
):
    def decorator(f):
        f.__cors__ = SimpleNamespace(
            _cors_origin=origin,
            _cors_expose_headers=expose_headers,
            _cors_supports_credentials=supports_credentials,
            _cors_allow_origins=(
                _parse_allow_origins(origin)
                if origin is not _default
                else origin
            ),
            _cors_allow_headers=(
                _parse_allow_headers(allow_headers)
                if allow_headers is not _default
                else allow_headers
            ),
            _cors_allow_methods=(
                _parse_allow_methods(allow_methods)
                if allow_methods is not _default
                else allow_methods
            ),
            _cors_max_age=(
                _parse_max_age(max_age) if max_age is not _default else max_age
            ),
        )
        return f

    return decorator


def _setup_cors_settings(app: Sanic) -> None:
    if app.config.CORS_ORIGINS == "*" and app.config.CORS_SUPPORTS_CREDENTIALS:
        raise SanicException(
            "Cannot use supports_credentials in conjunction with "
            "an origin string of '*'. See: "
            "http://www.w3.org/TR/cors/#resource-requests"
        )

    allow_headers = _get_allow_headers(app)
    allow_methods = _get_allow_methods(app)
    allow_origins = _get_allow_origins(app)
    expose_headers = _get_expose_headers(app)
    max_age = _get_max_age(app)

    app.ctx.cors = CORSSettings(
        allow_headers=allow_headers,
        allow_methods=allow_methods,
        allow_origins=allow_origins,
        always_send=app.config.CORS_ALWAYS_SEND,
        automatic_options=app.config.CORS_AUTOMATIC_OPTIONS,
        expose_headers=expose_headers,
        max_age=max_age,
        send_wildcard=(
            app.config.CORS_SEND_WILDCARD and WILDCARD_PATTERN in allow_origins
        ),
        supports_credentials=app.config.CORS_SUPPORTS_CREDENTIALS,
    )


def _get_from_cors_ctx(request: Request, key: str, default: Any = None):
    if request.route:
        value = getattr(request.route.ctx._cors, key, default)
        if value is not _default:
            return value
    return default


def _add_origin_header(request: Request, response: HTTPResponse) -> None:
    request_origin = request.headers.get("origin")
    origin_value = ""
    allow_origins = _get_from_cors_ctx(
        request,
        "_cors_allow_origins",
        request.app.ctx.cors.allow_origins,
    )
    fallback_origin = _get_from_cors_ctx(
        request,
        "_cors_origin",
        request.app.config.CORS_ORIGINS,
    )

    if request_origin:
        if request.app.ctx.cors.send_wildcard:
            origin_value = "*"
        else:
            for pattern in allow_origins:
                if pattern.match(request_origin):
                    origin_value = request_origin
                    break
    elif request.app.ctx.cors.always_send:
        if WILDCARD_PATTERN in allow_origins:
            origin_value = "*"
        else:
            if isinstance(fallback_origin, str) and "," not in fallback_origin:
                origin_value = fallback_origin
            else:
                origin_value = request.app.config.get("SERVER_NAME", "")

    if origin_value:
        response.headers[ORIGIN_HEADER] = origin_value


def _add_expose_header(request: Request, response: HTTPResponse) -> None:
    with_credentials = _is_request_with_credentials(request)
    headers = None
    expose_headers = _get_from_cors_ctx(
        request, "_cors_expose_headers", request.app.ctx.cors.expose_headers
    )
    # MDN: The value "*" only counts as a special wildcard value for requests
    # without credentials (requests without HTTP cookies or HTTP
    # authentication information). In requests with credentials, it is
    # treated as the literal header name "*" without special semantics.
    # Note that the Authorization header can't be wildcarded and always
    # needs to be listed explicitly.
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Expose-Headers
    if not with_credentials and "*" in expose_headers:
        headers = ["*"]
    elif expose_headers:
        headers = expose_headers

    if headers:
        response.headers[EXPOSE_HEADER] = ",".join(headers)


def _add_credentials_header(request: Request, response: HTTPResponse) -> None:
    supports_credentials = _get_from_cors_ctx(
        request,
        "_cors_supports_credentials",
        request.app.ctx.cors.supports_credentials,
    )
    if supports_credentials:
        response.headers[CREDENTIALS_HEADER] = "true"


def _add_allow_header(request: Request, response: HTTPResponse) -> None:
    with_credentials = _is_request_with_credentials(request)
    request_headers = set(
        h.strip().lower()
        for h in request.headers.get(REQUEST_HEADERS_HEADER, "").split(",")
    )
    allow_headers = _get_from_cors_ctx(
        request, "_cors_allow_headers", request.app.ctx.cors.allow_headers
    )

    # MDN: The value "*" only counts as a special wildcard value for requests
    # without credentials (requests without HTTP cookies or HTTP
    # authentication information). In requests with credentials,
    # it is treated as the literal header name "*" without special semantics.
    # Note that the Authorization header can't be wildcarded and always needs
    # to be listed explicitly.
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Headers
    if not with_credentials and "*" in allow_headers:
        allow_headers = ["*"]
    else:
        allow_headers = request_headers & allow_headers

    if allow_headers:
        response.headers[ALLOW_HEADERS_HEADER] = ",".join(allow_headers)


def _add_max_age_header(request: Request, response: HTTPResponse) -> None:
    max_age = _get_from_cors_ctx(
        request, "_cors_max_age", request.app.ctx.cors.max_age
    )
    if max_age:
        response.headers[MAX_AGE_HEADER] = max_age


def _add_methods_header(request: Request, response: HTTPResponse) -> None:
    # MDN: The value "*" only counts as a special wildcard value for requests
    # without credentials (requests without HTTP cookies or HTTP
    # authentication information). In requests with credentials, it
    # is treated as the literal method name "*" without
    # special semantics.
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Access-Control-Allow-Methods
    methods = None
    with_credentials = _is_request_with_credentials(request)
    allow_methods = _get_from_cors_ctx(
        request, "_cors_allow_methods", request.app.ctx.cors.allow_methods
    )

    if not with_credentials and "*" in allow_methods:
        methods = {"*"}
    elif request.route:
        group = request.app.router.groups.get(request.route.segments)
        if group:
            group_methods = {method.lower() for method in group.methods}
            if allow_methods:
                methods = group_methods & allow_methods
            else:
                methods = group_methods

    if methods:
        response.headers[ALLOW_METHODS_HEADER] = ",".join(methods).upper()


def _add_vary_header(request: Request, response: HTTPResponse) -> None:
    allow_origins = _get_from_cors_ctx(
        request,
        "_cors_allow_origins",
        request.app.ctx.cors.allow_origins,
    )
    if len(allow_origins) > 1:
        response.headers[VARY_HEADER] = "origin"


def _get_allow_origins(app: Sanic) -> Tuple[re.Pattern, ...]:
    origins = app.config.CORS_ORIGINS
    return _parse_allow_origins(origins)


def _parse_allow_origins(
    value: Union[str, re.Pattern, List[Union[str, re.Pattern]]],
) -> Tuple[re.Pattern, ...]:
    origins: Optional[
        Union[List[str], List[re.Pattern], List[Union[str, re.Pattern]]]
    ] = None
    if value and isinstance(value, str):
        if value == "*":
            origins = [WILDCARD_PATTERN]
        else:
            origins = [origin.strip() for origin in value.split(",")]
    elif isinstance(value, re.Pattern):
        origins = [value]
    elif isinstance(value, list):
        origins = value

    return tuple(
        pattern
        if isinstance(pattern, re.Pattern)
        else re.compile(re.escape(pattern))
        for pattern in (origins or [])
    )


def _get_expose_headers(app: Sanic) -> FrozenSet[str]:
    expose_headers = (
        (
            app.config.CORS_EXPOSE_HEADERS
            if isinstance(
                app.config.CORS_EXPOSE_HEADERS, (list, set, frozenset, tuple)
            )
            else app.config.CORS_EXPOSE_HEADERS.split(",")
        )
        if app.config.CORS_EXPOSE_HEADERS
        else tuple()
    )
    return frozenset(header.lower() for header in expose_headers)


def _get_allow_headers(app: Sanic) -> FrozenSet[str]:
    return _parse_allow_headers(app.config.CORS_ALLOW_HEADERS)


def _parse_allow_headers(value: str) -> FrozenSet[str]:
    allow_headers = (
        (
            value
            if isinstance(
                value,
                (list, set, frozenset, tuple),
            )
            else value.split(",")
        )
        if value
        else tuple()
    )
    return frozenset(header.lower() for header in allow_headers)


def _get_max_age(app: Sanic) -> str:
    return _parse_max_age(app.config.CORS_MAX_AGE or "")


def _parse_max_age(value) -> str:
    max_age = value or ""
    if isinstance(max_age, timedelta):
        max_age = str(int(max_age.total_seconds()))
    return str(max_age)


def _get_allow_methods(app: Sanic) -> FrozenSet[str]:
    return _parse_allow_methods(app.config.CORS_METHODS)


def _parse_allow_methods(value) -> FrozenSet[str]:
    allow_methods = (
        (
            value
            if isinstance(
                value,
                (list, set, frozenset, tuple),
            )
            else value.split(",")
        )
        if value
        else tuple()
    )
    return frozenset(method.lower() for method in allow_methods)


def _is_request_with_credentials(request: Request) -> bool:
    return bool(request.headers.get("authorization") or request.cookies)



from sanic import Sanic
from sanic.response import text

from sanic_ext import openapi

from .utils import get_spec


def test_func_handlers(app: Sanic):
    class TestClass:
        @staticmethod
        @openapi.definition(
            summary="staticmethod custom summary",
            tag="Test",
        )
        async def staticmethod_handler(request):
            return text("...")

        @classmethod
        @openapi.definition(
            summary="classmethod custom summary",
            tag="Test",
        )
        async def classmethod_handler(cls, request):
            return text("...")

        @openapi.definition(
            summary="instance method custom summary",
            tag="Test",
        )
        async def instance_method_handler(self, request):
            return text("...")

    app.add_route(TestClass.staticmethod_handler, "/staticmethod")
    app.add_route(TestClass.classmethod_handler, "/classmethod")
    app.add_route(TestClass().instance_method_handler, "/instance_method")

    spec = get_spec(app)
    paths = spec["paths"]
    assert len(paths) == 3
    assert (
        paths["/staticmethod"]["get"]["summary"]
        == "staticmethod custom summary"
    )
    assert (
        paths["/classmethod"]["get"]["summary"] == "classmethod custom summary"
    )
    assert (
        paths["/instance_method"]["get"]["summary"]
        == "instance method custom summary"
    )



from dataclasses import dataclass
from datetime import datetime
from typing import Dict

import attrs
import pytest
from msgspec import Struct
from pydantic import BaseModel
from pydantic.dataclasses import dataclass as pydataclass

from sanic_ext import openapi

from .utils import get_spec


@dataclass
class AlertDataclass:
    hit: Dict[str, int]
    last_updated: datetime


@dataclass
class AlertResponseDataclass:
    alert: AlertDataclass
    rule_id: str


class AlertPydanticBaseModel(BaseModel):
    hit: Dict[str, int]
    last_updated: datetime


class AlertResponsePydanticBaseModel(BaseModel):
    alert: AlertPydanticBaseModel
    rule_id: str


class AlertMsgspecBaseModel(Struct):
    hit: Dict[str, int]
    last_updated: datetime


class AlertResponseMsgspecBaseModel(Struct):
    alert: AlertMsgspecBaseModel
    rule_id: str


@pydataclass
class AlertPydanticDataclass:
    hit: Dict[str, int]
    last_updated: datetime


@pydataclass
class AlertResponsePydanticDataclass:
    alert: AlertPydanticDataclass
    rule_id: str


@attrs.define
class AlertAttrs:
    hit: Dict[str, int]
    last_updated: datetime


@attrs.define
class AlertResponseAttrs:
    alert: AlertAttrs
    rule_id: str


@pytest.mark.parametrize(
    "AlertResponse,check_alert",
    (
        (AlertResponseDataclass, False),
        (AlertResponseAttrs, False),
        (AlertResponseMsgspecBaseModel, True),
        (AlertResponsePydanticBaseModel, True),
        (AlertResponsePydanticDataclass, True),
    ),
)
def test_pydantic_base_model(app, AlertResponse, check_alert):
    @app.get("/")
    @openapi.definition(
        body={"application/json": openapi.Component(AlertResponse)}
    )
    async def handler(_):
        ...

    spec = get_spec(app)
    alert_response_name = AlertResponse.__name__
    alert_name = "Alert" + alert_response_name[13:]

    assert spec["paths"]["/"]["get"]["requestBody"] == {
        "content": {
            "application/json": {
                "schema": {
                    "$ref": f"#/components/schemas/{alert_response_name}"
                }
            }
        }
    }
    assert alert_response_name in spec["components"]["schemas"]

    if check_alert:
        assert alert_name in spec["components"]["schemas"]
        assert spec["components"]["schemas"][alert_response_name][
            "properties"
        ]["alert"] == {"$ref": f"#/components/schemas/{alert_name}"}



from dataclasses import dataclass

import pytest
from sanic import Sanic, json


@dataclass
class Foo:
    bar: int


def test_constant_in_registry(app: Sanic):
    assert len(app.ext._constant_registry._registry) == 0
    app.ext.add_constant("bar", 999)
    assert len(app.ext._constant_registry._registry) == 1
    assert "bar" in app.ext._constant_registry


def test_constant_is_injected(app: Sanic):
    app.ext.add_constant("bar", 999)

    @app.get("/")
    async def handler(_, bar: int):
        return json({"bar": bar})

    _, response = app.test_client.get("/")

    assert response.json["bar"] == 999
    assert app.config.BAR == 999


def test_constant_is_injected_into_constructor(app: Sanic):
    app.ext.add_constant("bar", 999)
    app.ext.add_dependency(Foo)

    @app.get("/")
    async def handler(_, foo: Foo):
        return json({"foo": foo.bar})

    _, response = app.test_client.get("/")

    assert response.json["foo"] == 999
    assert app.config.BAR == 999


def test_load_config(app: Sanic):
    app.ext.load_constants()
    assert len(app.ext._constant_registry._registry) == 0
    app.ext.load_constants({**app.config}, overwrite=True)
    assert len(app.ext._constant_registry._registry) == len(
        [k for k in app.config.keys() if k.isupper()]
    )
    assert "request_buffer_size" in app.ext._constant_registry


def test_load_custom(app: Sanic):
    app.ext.load_constants({"foo": "bar"})
    assert len(app.ext._constant_registry._registry) == 1
    assert app.ext._constant_registry.get("foo") == "bar"


def test_load_overwrite(app: Sanic):
    app.ext.load_constants({"foo": "bar"})
    with pytest.raises(
        ValueError, match="A value for FOO has already been assigned"
    ):
        app.ext.load_constants({"foo": "bar"})
    app.ext.load_constants({"foo": "bar"}, True)



from sanic import Request, Sanic
from sanic.response import text

from sanic_ext import openapi
from sanic_ext.extensions.openapi.definitions import ExternalDocumentation

from .utils import get_spec


def test_external_docs(app: Sanic):
    @app.route("/test0")
    @openapi.document("http://example.com/more", "Find more info here")
    async def handler0(request: Request):
        return text("ok")

    @app.route("/test1")
    @openapi.definition(
        document=ExternalDocumentation(
            "http://example.com/more", "Find more info here"
        )
    )
    async def handler1(request: Request):
        return text("ok")

    @app.route("/test2")
    @openapi.definition(document="http://example.com/more")
    async def handler2(request: Request):
        return text("ok")

    @app.route("/test3")
    async def handler3(request: Request):
        """
        openapi:
        ---
        summary: This is a summary.
        externalDocs:
          description: Find more info here
          url: http://example.com/more
        """
        return text("ok")

    @app.route("/test4")
    @openapi.document(
        ExternalDocumentation("http://example.com/more", "Find more info here")
    )
    async def handler4(request: Request):
        return text("ok")

    spec = get_spec(app)
    paths = spec["paths"]
    assert len(paths) == 5
    for i in range(5):
        doc_obj = paths[f"/test{i}"]["get"]["externalDocs"]
        assert doc_obj["url"] == "http://example.com/more"
        if i != 2:
            assert doc_obj["description"] == "Find more info here"



from sys import version_info
from typing import Literal, Union

import pytest
from sanic import text

from sanic_ext import validate


def _dataclass_spec(annotation):
    from dataclasses import dataclass

    @dataclass
    class Spec:
        name: annotation

    return Spec


def _attrs_spec(annotation):
    import attrs

    @attrs.define
    class Spec:
        name: annotation

    return Spec


def _msgspec_spec(annotation):
    from msgspec import Struct

    class Spec(Struct):
        name: annotation

    return Spec


def _pydantic_spec(annotation):
    from pydantic.dataclasses import dataclass as pydataclass

    @pydataclass
    class Spec:
        name: annotation

    return Spec


@pytest.mark.parametrize(
    "annotation",
    (
        (
            Literal["foo"],
            Literal["foo", "bar"],
            Union[Literal["foo"], Literal["bar"]],
        )
    ),
)
@pytest.mark.parametrize(
    "spec_builder",
    (
        _dataclass_spec,
        _attrs_spec,
        _msgspec_spec,
        _pydantic_spec,
    ),
)
def test_literal(app, annotation, spec_builder):
    Spec = spec_builder(annotation)

    @app.get("/")
    @validate(query=Spec)
    def route(_, query: Spec):
        return text(query.name)

    _, response = app.test_client.get("", params={"name": "foo"})
    assert response.text == "foo"


@pytest.mark.skipif(version_info < (3, 10), reason="Not needed on 3.10")
@pytest.mark.parametrize(
    "spec_builder",
    (
        _dataclass_spec,
        _attrs_spec,
        _msgspec_spec,
        _pydantic_spec,
    ),
)
def test_literal_3_10(app, spec_builder):
    Spec = spec_builder(Literal["foo"] | Literal["bar"])

    @app.get("/")
    @validate(query=Spec)
    def route(_, query: Spec):
        return text(query.name)

    _, response = app.test_client.get("", params={"name": "foo"})
    assert response.text == "foo"



from sanic.exceptions import SanicException


class ValidationError(SanicException):
    status_code = 400


class InitError(SanicException):
    ...


class ExtensionNotFound(SanicException):
    ...



import types
import typing

from inspect import isclass


try:
    UnionType = types.UnionType  # type: ignore
except AttributeError:
    UnionType = type("UnionType", (), {})  # type: ignore

try:
    from pydantic import BaseModel

    PYDANTIC = True
except ImportError:
    PYDANTIC = False

try:
    import attrs  # noqa

    ATTRS = True
except ImportError:
    ATTRS = False

try:
    from msgspec import Struct

    MSGSPEC = True
except ImportError:
    MSGSPEC = False


def is_generic(item):
    return (
        isinstance(item, typing._GenericAlias)
        or isinstance(item, UnionType)
        or hasattr(item, "__origin__")
    )


def is_optional(item):
    if is_generic(item):
        args = typing.get_args(item)
        return len(args) == 2 and type(None) in args
    return False


def is_pydantic(model):
    return PYDANTIC and (
        issubclass(model, BaseModel) or hasattr(model, "__pydantic_model__")
    )


def is_attrs(model):
    return ATTRS and (hasattr(model, "__attrs_attrs__"))


def is_msgspec(model):
    return MSGSPEC and issubclass(model, Struct)


def flat_values(
    item: typing.Union[
        typing.Dict[str, typing.Any], typing.Iterable[typing.Any]
    ],
) -> typing.Set[typing.Any]:
    values = set()
    if isinstance(item, dict):
        item = item.values()
    for value in item:
        if isinstance(value, dict) or isinstance(value, list):
            values.update(flat_values(value))
        else:
            values.add(value)
    return values


def contains_annotations(d: typing.Dict[str, typing.Any]) -> bool:
    values = flat_values(d)
    return any(isclass(q) or is_generic(q) for q in values)



from itertools import count

from sanic import Request, Sanic
from sanic.compat import Header
from sanic.models.protocol_types import TransportProtocol


class CountedRequest(Request):
    __slots__ = ()

    _counter = count()
    count = next(_counter)

    def __init__(
        self,
        url_bytes: bytes,
        headers: Header,
        version: str,
        method: str,
        transport: TransportProtocol,
        app: Sanic,
        head: bytes = b"",
        stream_id: int = 0,
    ):
        super().__init__(
            url_bytes,
            headers,
            version,
            method,
            transport,
            app,
            head,
            stream_id,
        )
        self.__class__._increment()
        if hasattr(self.app, "multiplexer"):
            self.app.multiplexer.state["request_count"] = self.__class__.count

    @classmethod
    def _increment(cls):
        cls.count = next(cls._counter)

    @classmethod
    def reset_count(cls):
        cls._counter = count()
        cls.count = next(cls._counter)
